#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Python Jamf Classic API
Author: Sam Forester
Contributors:
James Reynolds (reynolds@biology.utah.edu)
Tony Williams (tonyw@honestpuck.com) (ARW)
"""

__author__ = "Sam Forester"
__email__ = "sam.forester@utah.edu"
__copyright__ = "Copyright (c) 2020 University of Utah, Marriott Library"
__license__ = "MIT"
__version__ = "0.5.2"

import json
import logging

import requests

from . import config, convert, exceptions

LOGLEVEL = logging.INFO


class Singleton(type):
    """allows us to share a single object"""

    _instances = {}

    def __call__(cls, *a, **kw):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*a, **kw)
        return cls._instances[cls]


class API(metaclass=Singleton):
    """
    Class for making api calls to JSS
    """

    session = False

    def __init__(
        self, config_path=None, hostname=None, username=None, password=None, prompt=True
    ):
        """
        Create requests.Session with JSS address and authentication

        :param config_path <str>:    Path to file containing config. If it
                                     starts with '~' character we pass it
                                     to expanduser.
        :param hostname <str>:       Hostname of server
        :param username <str>:       username for server
        :param password <str>:       password for server
        :param prompt <bool>:        Allow the script to prompt if any info is missing
        """
        self.log = logging.getLogger(f"{__name__}.API")
        self.log.setLevel(LOGLEVEL)
        # Load Prefs and Init session
        self.config = config.Config(
            config_path=config_path,
            hostname=hostname,
            username=username,
            password=password,
            prompt=prompt,
        )
        hostname = hostname or self.config.hostname
        self.username = username or self.config.username
        self.password = password or self.config.password
        if self.config.password:
            self.save_token_in_keyring = True
        else:
            self.save_token_in_keyring = False

        if not hostname and not self.username and not self.password:
            raise exceptions.JamfConfigError(
                "No jamf hostname or credentials could be found."
            )
        while hostname[-1] == "/":
            hostname = hostname[:-1]
        self.hostname = hostname
        self.session = requests.Session()

    def get_token(self, old_token=None):
        with requests.Session() as session:
            if old_token:
                session.headers.update({"Authorization": f"Bearer {old_token}"})
                url = f"{self.hostname}/api/v1/auth/keep-alive"
            else:
                session.auth = (self.username, self.password)
                url = f"{self.hostname}/api/v1/auth/token"
            response = self._submit_request(session, "post", url)
            if response.status_code != 200:
                print("Server did not return a bearer token")
                return None
            try:
                json_data = json.loads(response.text)
            except Exception:
                print("Couldn't parse bearer token json")
                return None
            if self.config.password:
                self.config.save_new_token(json_data["token"], json_data["expires"])
            return json_data["token"]

    def revoke_token(self):
        self.config.revoke_token()

    def _set_session_auth(self):
        """set the session Jamf Pro API token or basic auth"""
        token = None
        # Check for old token and renew it if found
        if self.config.password:
            self.config.load_token()
            if self.config.expired:
                token = self.get_token(old_token=self.config.token)
        # Get a new token
        if not token:
            token = self.get_token()
        # Only use token if jamf version is >= 10.35.0
        use_token = False
        if token:
            with requests.Session() as session:
                url = f"{self.hostname}/api/v1/jamf-pro-version"
                session.headers.update({"Authorization": f"Bearer {token}"})
                response = self._submit_request(session, "get", url)
                if response.status_code == 200:
                    try:
                        json_data = json.loads(response.text)
                        version = json_data["version"].split("-")[0]
                        v1 = tuple(map(int, (version.split("."))))
                        v2 = tuple(map(int, (10, 35, 0)))
                        if v1 >= v2:
                            use_token = True
                    except Exception:
                        print("Couldn't parse jamf version json")
        if use_token:
            self.session.headers.update({"Authorization": f"Bearer {token}"})
        else:
            self.session.auth = (self.username, self.password)

    def _submit_request(self, session, method, url, data=None, raw=False):
        """
        Generic request

        :param method <str>:   get | post | put | delete
        :param url <str>: API url (e.g. "http://server/JSSResource/policies/id/1")
        :param data <dict>:    data submitted (get and delete ignore this)
        :param raw <bool>:     return requests.Response obj  (skip errors)

        :returns <dict|requests.Response>:
        """
        self.log.debug(f"{method}: {url}")
        if isinstance(data, dict):
            data = convert.dict_to_xml(data)
            data = data.encode("utf-8")
            self.log.debug("xml data: %s", data)
        session_method = getattr(session, method)
        try:
            response = session_method(url, data=data)
        except requests.exceptions.ConnectionError as error:
            raise exceptions.JamfNoConnectionError(
                f"Could not connect to {self.hostname}\n{error}"
            )
        if raw:
            return response
        if not response.ok:
            if response.status_code == 401:
                raise exceptions.JamfAuthenticationError(response, "Bad auth")
            elif response.status_code == 404:
                raise exceptions.JamfRecordNotFound(response, "Not Found")
            else:
                raise exceptions.JamfConnectionError(
                    response, f"Error: {response.status_code}"
                )
        self.log.debug("response.text: %s", response.text)
        # return successful response data (usually: {'id': jssid})
        return response

    def _crud(self, method, endpoint, data=None, raw=False):
        self._set_session_auth()
        self.session.headers.update({"Accept": "application/xml"})
        url = f"{self.hostname}/JSSResource/{endpoint}"
        response = self._submit_request(self.session, method, url, data, raw)
        convert_data = convert.xml_to_dict(response.text)
        self.log.debug("converted data: %s", convert_data)
        return convert_data

    def get(self, endpoint, raw=False):
        """
        Get JSS information

        :param endpoint <str>:  API endpoint (e.g. "policies/id/1")
        :param raw <bool>:      return requests.Response obj  (skip errors)

        :returns <dict|requests.Response>:
        """
        return self._crud("get", endpoint, None, raw)

    def post(self, endpoint, data, raw=False):
        """
        Create new entries on JSS

        :param endpoint <str>:  JSS endpoint (e.g. "policies/id/0")
        :param data <dict>:     data to be submitted
        :param raw <bool>:      return requests.Response obj  (skip errors)

        :returns dict:          response data
        """
        return self._crud("post", endpoint, data, raw)

    def put(self, endpoint, data, raw=False):
        """
        Update existing entries on JSS

        :param endpoint <str>:  JSS endpoint (e.g. "policies/id/0")
        :param data <dict>:     data to be submitted
        :param raw <bool>:      return requests.Response obj  (skip errors)

        :returns dict:          response data
        """
        return self._crud("put", endpoint, data, raw)

    def delete(self, endpoint, raw=False):
        """
        Delete entry on JSS

        :param endpoint <str>:  API endpoint (e.g. "policies/id/1")
        :param raw  <bool>:     return requests.Response obj (skip errors)

        :returns <dict|requests.Response>:
        """
        return self._crud("delete", endpoint, None, raw)

    def __del__(self):
        if self.session:
            self.log.debug("closing session")
            self.session.close()



# -*- coding: utf-8 -*-

"""
Tests for jamf.api
"""

__author__ = "James Reynolds"
__email__ = "reynolds@biology.utah.edu"
__copyright__ = "Copyright (c) 2022 University of Utah"
__license__ = "MIT"
__version__ = "0.1.0"

import logging
import os
import random
import string
import unittest

from jamf import api, exceptions

HOSTNAME = "http://localhost"
USERNAME = "python-jamf"
PASSWORD = "secret"

BAD_HOSTNAME = "http://asdfasdf"
BAD_HOSTNAME_PORT = "http://github:80"
BAD_USERNAME = "asdf"


def get_creds():
    if "JAMF_HOSTNAME" in os.environ:
        hostname = os.environ["JAMF_HOSTNAME"]
    else:
        hostname = HOSTNAME
    if "JAMF_USERNAME" in os.environ:
        username = os.environ["JAMF_USERNAME"]
    else:
        username = USERNAME
    if "JAMF_PASSWORD" in os.environ:
        password = os.environ["JAMF_PASSWORD"]
    else:
        password = PASSWORD
    return (hostname, username, password)


class TestHost(unittest.TestCase):
    def test_successful_connection(self):
        """
        Test successful connection
        """
        api.API._instances = {}
        hostname, username, password = get_creds()
        server = api.API(hostname=hostname, username=username, password=password)
        server.revoke_token()
        accounts = server.get("accounts")
        self.assertTrue("accounts" in accounts)

    def test_bad_prefs(self):
        """
        Test bad prefs
        """
        api.API._instances = {}
        self.assertRaises(
            exceptions.JamfConfigError,
            lambda: api.API(
                config_path="/var/false",
                hostname="",
                username="",
                password="",
                prompt=False,
            ),
        )

    def test_bad_hostname(self):
        """
        Test bad hostname
        """
        api.API._instances = {}
        hostname, username, password = get_creds()
        server = api.API(hostname=BAD_HOSTNAME, username=username, password=password)
        server.revoke_token()
        self.assertRaises(
            exceptions.JamfNoConnectionError, lambda: server.get("accounts")
        )

    def test_bad_port(self):
        """
        Test bad port
        """
        api.API._instances = {}
        hostname, username, password = get_creds()
        server = api.API(
            hostname=BAD_HOSTNAME_PORT, username=username, password=password
        )
        server.revoke_token()
        self.assertRaises(
            exceptions.JamfNoConnectionError, lambda: server.get("accounts")
        )

    def test_bad_username_password(self):
        """
        Test bad username password
        """
        api.API._instances = {}
        hostname, username, password = get_creds()
        server = api.API(hostname=hostname, username=BAD_USERNAME, password=password)
        server.revoke_token()
        self.assertRaises(
            exceptions.JamfAuthenticationError, lambda: server.get("accounts")
        )


class TestAPI(unittest.TestCase):
    def setUp(self):
        api.API._instances = {}
        hostname, username, password = get_creds()
        self.server = api.API(hostname=hostname, username=username, password=password)

    def test_get_token(self):
        """
        Test get_token
        """

        token = self.server.get_token()
        self.assertTrue(len(token) == 331)

    def test_revoke_token(self):
        """
        Test revoke_token
        """
        token1 = self.server.get_token()
        self.server.revoke_token()
        token2 = self.server.get_token()
        self.assertTrue(token1 != token2)

    def test_get_accounts(self):
        """
        Test get accounts
        """
        accounts = self.server.get("accounts")
        self.assertTrue("accounts" in accounts)

    def test_get_account_1(self):
        """
        Test get accounts
        """
        account = self.server.get("accounts/userid/1")
        self.assertTrue("account" in account)
        self.assertTrue("password_sha256" in account["account"])

    def test_crud(self):
        """
        Test CRUD: post new record, get, assert email, change email, put, get, assert email, delete, assert not found
        """
        name = "python-jamf-" + "".join(random.choices(string.ascii_letters, k=16))
        data = {
            "account": {
                "access_level": "Full Access",
                "directory_user": "false",
                "email": "python@jamf.jamf",
                "email_address": "python@jamf.jamf",
                "enabled": "Disabled",
                "force_password_change": "false",
                "full_name": "Python-Jamf",
                "name": name,
                "password": "probably-not-a-good-test",
                "privilege_set": "Custom",
                "privileges": {
                    "casper_admin": None,
                    "casper_remote": None,
                    "jss_actions": None,
                    "jss_objects": None,
                    "jss_settings": None,
                    "recon": None,
                },
            }
        }
        self.server.post("accounts/userid/0", data)
        account = self.server.get(f"accounts/username/{name}")
        self.assertTrue(account["account"]["email"] == "python@jamf.jamf")
        account["account"]["email"] = "test@test.test"
        self.server.put(f"accounts/userid/{account['account']['id']}", account)
        account = self.server.get(f"accounts/username/{name}")
        self.assertTrue(account["account"]["email"] == "test@test.test")
        self.server.delete(f"accounts/username/{name}")
        self.assertRaises(
            exceptions.JamfRecordNotFound,
            lambda: self.server.get(f"accounts/username/{name}"),
        )


if __name__ == "__main__":
    FMT = "%(asctime)s: %(levelname)8s: %(name)s - %(funcName)s(): %(message)s"
    logging.basicConfig(level=logging.DEBUG, format=FMT)
    unittest.main(verbosity=1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Record and Records

Combination of Sam Forester's Category class and Tony Williams' Records class
with lots of gluing, fleshing out, and other improvements by James Reynolds

James Reynolds reynolds@biology.utah.edu
Sam Forester sam.forester@utah.edu
Tony Williams tonyw@honestpuck.com

Copyright (c) 2020 University of Utah, Marriott Library
Copyright (c) 2020 Tony Williams
"""

__author__ = "James Reynolds, Sam Forester, Tony Williams"
__email__ = "reynolds@biology.utah.edu, sam.forester@utah.edu, tonyw@honestpuck.com"
__copyright__ = "Copyright (c) 2021 University of Utah, School of Biological Sciences and Copyright (c) 2020 Tony Williams"
__license__ = "MIT"
__date__ = "2020-09-21"
__version__ = "0.4.6"


import copy
import json
import logging
import os.path
import re
from pprint import pprint
from sys import stderr

# pylint: disable=relative-beyond-top-level
from .api import API

__all__ = (
    "AdvancedComputerSearches",
    "AdvancedMobileDeviceSearches",
    "AdvancedUserSearches",
    "Buildings",
    "BYOProfiles",
    "Categories",
    "Classes",
    "ComputerConfigurations",
    "ComputerExtensionAttributes",
    "ComputerGroups",
    "ComputerReports",
    "Computers",
    "Departments",
    "DirectoryBindings",
    "DiskEncryptionConfigurations",
    "DistributionPoints",
    "DockItems",
    "Ebooks",
    "Ibeacons",
    "JSONWebTokenConfigurations",
    "LDAPServers",
    "MacApplications",
    "ManagedPreferenceProfiles",
    "MobileDeviceApplications",
    "MobileDeviceCommands",
    "MobileDeviceConfigurationProfiles",
    "MobileDeviceEnrollmentProfiles",
    "MobileDeviceExtensionAttributes",
    "MobileDeviceInvitations",
    "MobileDeviceProvisioningProfiles",
    "MobileDevices",
    "NetbootServers",
    "NetworkSegments",
    "OSXConfigurationProfiles",
    "Packages",
    "PatchExternalSources",
    "PatchInternalSources",
    "PatchPolicies",
    "PatchSoftwareTitles",
    "Peripherals",
    "PeripheralTypes",
    "Policies",
    "Printers",
    "RemovableMACAddresses",
    "Scripts",
    "Sites",
    "SoftwareUpdateServers",
    "UserExtensionAttributes",
    "UserGroups",
    "Users",
    "VPPAccounts",
    "VPPAssignments",
    "VPPInvitations",
    "WebHooks",
    # Add all non-Jamf Record classes to the exclude list in valid_records below
    "JamfError",
)


def valid_records():
    valid = tuple(
        x
        for x in __all__
        if x
        not in [
            # Exclude list, add all non-Jamf Record classes here
            "JamfError"
        ]
    )
    return valid


# pylint: disable=eval-used
def class_name(name, case_sensitive=True):
    if case_sensitive and name in valid_records():
        return eval(name)
    if not case_sensitive:
        for temp in valid_records():
            if name.lower() == temp.lower():
                return eval(temp)
    raise JamfError(f"{name} is not a valid record.")


# pylint: disable=super-init-not-called
class JamfError(Exception):
    def __init__(self, message):
        self.message = message


# pylint: disable=super-init-not-called
class NotFound(Exception):
    pass


class Singleton(type):
    """allows us to share a single object"""

    _instances = {}

    def __call__(cls, *a, **kw):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*a, **kw)
        return cls._instances[cls]


class ClassicSwagger(metaclass=Singleton):
    def __init__(self):
        self._swagger = json.load(
            open(os.path.dirname(__file__) + "/records.json", "r")
        )
        self._broken_api = []
        post_template2 = {"general": {"name": "%NAME%"}}
        self._post_templates = {
            "BYOProfiles": post_template2,
            "ComputerConfigurations": post_template2,
            "ComputerReports": post_template2,
            "Computers": post_template2,
            "DirectoryBindings": post_template2,
            "Ebooks": post_template2,
            "JSONWebTokenConfigurations": post_template2,
            # "LicensedSoftware": post_template2,
            # "LDAPServers": post_template2,
            "MacApplications": post_template2,
            "ManagedPreferenceProfiles": post_template2,
            "MobileDevices": post_template2,
            "MobileDeviceApplications": post_template2,
            "MobileDeviceConfigurationProfiles": post_template2,
            "MobileDeviceEnrollmentProfiles": post_template2,
            "MobileDeviceProvisioningProfiles": post_template2,
            "OSXConfigurationProfiles": post_template2,
            "Peripherals": post_template2,
            "Policies": post_template2,
            "SoftwareUpdateServers": post_template2,
            "VPPAccounts": post_template2,
            "VPPAssignments": post_template2,
            "VPPInvitations": post_template2,
            "ComputerGroups": {"name": "%NAME%", "is_smart": True},
            "DistributionPoints": {
                "name": "%NAME%",
                "read_only_username": "read_only_username",
                "read_write_username": "read_write_username",
                "share_name": "files",
            },
            "DockItems": {
                "name": "%NAME%",
                "path": "file://localhost/Applications/Safari.app/",
                "type": "App",
            },
            "Ibeacons": {
                "name": "%NAME%",
                "uuid": "7710B6A4-FD29-4647-B2F4-B3FA645146A8",  # I don't have iBeacons, I don't know the purpose of this value, I got it with `uuidgen` (https://support.twocanoes.com/hc/en-us/articles/203081205-Managing-Printers-with-iBeacons)
            },
            "NetbootServers": {"name": "%NAME%", "ip_address": "10.0.0.1"},
            "NetworkSegments": {
                "name": "%NAME%",
                "starting_address": "10.0.0.1",
                "ending_address": "10.0.0.1",
            },
            "Packages": {
                "name": "%NAME%",
                "filename": "filename.pkg",
            },
            "PatchExternalSources": {
                "name": "%NAME%",
                "host_name": "example.com",
            },
            "PatchPolicies": {
                "general": {"name": "%NAME%", "target_version": "%VERSION%"}
            },
            "PatchSoftwareTitles": {"name_id": "%NAME%", "source_id": "1"},
            # "RestrictedSoftware": {"general":{"name":"%NAME%","process_name":"%NAME%"}},
            "UserGroups": {"general": {"name": "%NAME%"}, "is_smart": False},
            "WebHooks": {
                "event": "ComputerAdded",
                "name": "%NAME%",
                "url": "http:/example.com",
            },
        }
        self._swagger_fixes = {
            "ComputerConfigurations": {
                "s2": "configuration",
            },
            "ComputerReports": {
                "s1": "computer_reports",
            },
            "MobileDeviceCommands": {
                "id_text1": "uuid",
                "id_text2": "uuid",
            },
            "RestrictedSoftware": {
                "p3": "restricted_software_title",
                "s1": "restricted_software",
            },
        }

    def post_template(self, className, name):
        if className in self._post_templates:
            template = copy.deepcopy(self._post_templates[className])
            if "name" in template:
                template["name"] = template["name"].replace("%NAME%", name)
            elif "name_id" in template:
                template["name_id"] = template["name_id"].replace("%NAME%", name)
            elif "general" in template and "name" in template["general"]:
                t = template["general"]["name"]
                template["general"]["name"] = t.replace("%NAME%", name)
        else:
            template = {"name": name}
        return template

    def swagger(self, cls, kk):
        fixes = {}
        if cls.__name__ in self._swagger_fixes:
            fixes = self._swagger_fixes[cls.__name__]

        # The endpoint url, e.g. "Policies" class becomes "policies" endpoint
        if "p1" in fixes:
            p1 = fixes["p1"]
        else:
            p1 = cls.__name__.lower()
        if kk == "path_name":
            return p1

        # Get the definition name, which almost always is the plural name
        # exceptions: LicensedSoftware, RestrictedSoftware?
        if "p2" in fixes:
            p2 = fixes["p2"]
        else:
            p2 = self.get_schema(p1)
            # If there's an xml entry, use it for the definition name
            temp2 = self._swagger["definitions"][p2]
            if "xml" in temp2 and "name" in temp2["xml"]:
                p2 = temp2["xml"]["name"]
        if kk == "def_name":
            return p2

        if "id_text1" in fixes:
            id1 = fixes["id_text1"]
        else:
            id1 = "id"
        if kk == "id1":
            return id1

        if kk == "p1, id1":
            return p1, id1

        end = f"{p1}/{id1}/"

        if kk == "end":
            return end

        if "id_text2" in fixes:
            id2 = fixes["id_text2"]
        else:
            id2 = "id"
        if kk == "id2":
            return id2

        # Get the schema, which almost always is the singular name
        if "p3" in fixes:
            p3 = fixes["p3"]
        else:
            temp1 = f"{p1}/{id1}/{{{id2}}}"
            p3 = self.get_schema(temp1)
        if kk == "p3":
            return p3

        if kk == "p1, p2, id1, p3":
            return p1, p2, id1, p3

        # Singular, which almost always is the p3
        if "s1" in fixes:
            s1 = fixes["s1"]
        else:
            s1 = p3
        if kk == "s1":
            return s1

        # This is the name of the endpoint when it's returned from a post
        # e.g. ComputerConfigurations: {'configuration': {'general': {'name': 'rfwlkzis'}}}
        if "s2" in fixes:
            s2 = fixes["s2"]
        else:
            s2 = s1
        if kk == "s2":
            return s2

        if kk == "s1, end":
            return s1, end

        if kk == "s1, s2, end":
            return s1, s2, end

    def get_schema(self, swagger_path):
        temp1 = self._swagger["paths"]["/" + swagger_path]["get"]
        schema = temp1["responses"]["200"]["schema"]["$ref"]
        if schema.startswith("#/definitions/"):
            schema = schema[14:]
        return schema

    def is_action_valid(self, className, action):
        p1, id1 = self.swagger(className, "p1, id1")
        p = f"/{p1}/{id1}/{{{id1}}}"
        if className == PatchPolicies and action == "post":
            p = "/patchpolicies/softwaretitleconfig/id/{softwaretitleconfigid}"
        if p in self._broken_api:
            return False
        return p in self._swagger["paths"] and action in self._swagger["paths"][p]


class Record:
    """
    A class for an object on Jamf Pro

    NOTE: For reasons known only to itself Jamf uses 'wordstogether' for the
    endpoint in the URL but 'underscore_between' for the XML tags in some
    endpoints and there are cases where the endpoint and object tag are more
    different than that.

    This means we need to know 3 strings for each object type, the endpoint,
    the top of the list, and the top of the object.

    Just in case that's not confusing enough the id tag is not always 'id'.
    """

    def __new__(cls, *args, **kw):
        """
        returns existing record if one has been instantiated
        """
        jamf_id = int(args[0])
        if not hasattr(cls, "_instances"):
            cls._instances = {}
        swag = ClassicSwagger()
        plural = eval(cls.plural_class)
        api = API()
        if jamf_id == 0:
            if not swag.is_action_valid(plural, "post"):
                print(
                    f"Creating a new record with an id of 0 causes a post, "
                    f"which isn't a valid action for the {cls.plural_class} "
                    f"record type."
                )
                return None
            s1, s2, end = swag.swagger(plural, "s1, s2, end")
            use_template = True
            if type(args[1]) is list:
                if type(args[1][0]) is str:
                    name = args[1][0]
                elif type(args[1][0]) is dict:
                    name = ""
                    use_template = False
            elif type(args[1]) is str or args[1] is None:
                name = args[1]
            if use_template:
                if cls.plural_class == "PatchPolicies":
                    if len(args[1]) < 3:
                        raise JamfError(
                            "patchpolicies requires 3 args to create records"
                        )
                    softwaretitleconfigid = args[1][1]
                    end = (
                        f"patchpolicies/softwaretitleconfig/id/{softwaretitleconfigid}"
                    )
                    out = {s1: swag.post_template(cls.plural_class, name)}
                    t = out["patch_policy"]["general"]["target_version"]
                    out["patch_policy"]["general"]["target_version"] = t.replace(
                        "%VERSION%", args[1][2]
                    )
                else:
                    end = f"{end}0"
                    out = {s1: swag.post_template(cls.plural_class, name)}
                _data = api.post(end, out)
            else:
                end = f"{end}0"
                _data = api.post(end, args[1][0])
            jamf_id = int(_data[s2]["id"])
        else:
            _data = {}
            try:
                name = args[1]
            except NameError:
                name = ""
        if jamf_id not in cls._instances:
            rec = super(Record, cls).__new__(cls)
            rec.cls = cls
            rec.plural = plural
            cls._instances[jamf_id] = rec
        rec.api = api
        rec = cls._instances[jamf_id]
        rec.id = int(jamf_id)
        rec._data = _data
        rec.s = swag
        rec.name = name
        return rec

    def __eq__(self, x):
        if isinstance(x, Record):
            return self is x
            # return self.name == x.name and self.id == x.id
        elif isinstance(x, int):
            return self.id == x
        elif isinstance(x, str):
            if x.isdigit() or x == "-1":
                return self.id == int(x)
            else:
                return self.name == x
        elif isinstance(x, dict):
            jamf_id = int(x.get("id", -1))
            return self.name == x.get("name") and self.id == jamf_id
        else:
            raise TypeError(f"can't test equality of {x!r}")

    def __lt__(self, x):
        if self.name and x.name and self.name < x.name:
            return True
        return False

    def __str__(self):
        if self.name is not None:
            return self.name
        else:
            return ""

    def __repr__(self):
        return f"{self.__class__.__name__}({self.id}, {self.name!r})"

    def refresh(self):
        s1, end = self.s.swagger(self.plural, "s1, end")
        end = f"{end}{self.id}"
        if not self.s.is_action_valid(self.plural, "get"):
            raise JamfError(f"get({end}) is an invalid action for get")
        results = self.api.get(end)
        if s1 not in results:
            print("---------------------------------------------\nData dump\n")
            pprint(results)
            raise JamfError(f"Endpoint {end} has no member named {s1} (s1).")
        if results[s1]:
            self._data = results[s1]
        else:
            self._data = {}

    def delete(self):
        end = self.s.swagger(self.plural, "end")
        end = f"{end}{self.id}"
        if not self.s.is_action_valid(self.plural, "delete"):
            raise JamfError(f"{end} is an invalid endpoint for delete")
        return self.api.delete(end)

    def save(self):
        s1, end = self.s.swagger(self.plural, "s1, end")
        end = f"{end}{self.id}"
        if not self.s.is_action_valid(self.plural, "put"):
            raise JamfError(f"{end} is an invalid endpoint for put")
        out = {s1: self._data}
        return self.api.put(end, out)

    @property
    def data(self):
        if not self._data:
            self.refresh()
        return self._data

    def get_path_worker(self, path, placeholder, index=0):
        current = path[index]
        if type(placeholder) is dict:
            if current in placeholder:
                if index + 1 >= len(path):
                    return placeholder[current]
                placeholder = self.get_path_worker(
                    path, placeholder[current], index + 1
                )
            else:
                raise NotFound
        elif type(placeholder) is list:
            # I'm not sure this is the best way to handle arrays...
            result = []
            for item in placeholder:
                if current in item:
                    if index + 1 < len(path):
                        result.append(
                            self.get_path_worker(path, item[current], index + 1)
                        )
                    else:
                        result.append(item[current])
            placeholder = result
        return placeholder

    def get_path(self, path):
        if not self._data:
            self.refresh()
        result = self.get_path_worker(path.split("/"), self._data)
        return result

    def force_array(self, parent, child_name):
        _data = parent[child_name]
        if "size" in parent:
            if int(parent["size"]) > 0:
                return _data
        else:
            if type(_data) is list:
                return _data
        return [_data]

    def set_path(self, path, value):
        temp1 = path.split("/")
        endpoint = temp1.pop()
        temp2 = "/".join(temp1)
        if len(temp2) > 0:
            placeholder = self.get_path(temp2)
        else:
            if not self._data:
                self.refresh()
            placeholder = self._data
        if placeholder:
            if endpoint in placeholder:
                placeholder[endpoint] = value
                return True
            else:
                print(f"Error: '{endpoint}' missing from ")
                pprint(placeholder)
                return False
        else:
            print("Error: empty data:")
            pprint(placeholder)
            return False


class RecordsIterator:
    def __init__(self, records):
        self._records = records
        self._ids = records.ids()
        self._index = 0

    def __next__(self):
        if self._index < (len(self._ids)):
            if self._index < len(self._ids):
                id = self._ids[self._index]
                result = self._records.recordWithId(id)
            self._index += 1
            return result
        # End of Iteration
        raise StopIteration


class Records:
    """
    A class for a list of objects on Jamf Pro

    NOTE: For reasons known only to itself Jamf uses 'wordstogether' for the
    endpoint in the URL but 'underscore_between' for the XML tags in some
    endpoints and there are cases where the endpoint and object tag are more
    different than that.

    This means we need to know 3 strings for each object type, the endpoint,
    the top of the list, and the top of the object.

    Just in case that's not confusing enough the id tag is not always 'id'.
    """

    def __new__(cls, *a, **kw):
        rec = super().__new__(cls)
        rec._records = {}
        rec.cls = cls
        rec.s = ClassicSwagger()
        return rec

    def __init__(self, api=None):
        self.log = logging.getLogger(f"{__name__}.Records")
        self.api = api or API()

        self.data = {}
        self._jamf_ids = {v.id: v for v in self._records.values()}
        self._names = {v.name: v for v in self._records.values()}

    def __iter__(self):
        return RecordsIterator(self)

    def __getitem__(self, item):
        return self.recordWithId(self.ids()[item])

    def __contains__(self, x):
        return True if self.find(x) else False

    def names(self):
        if not self.data:
            self.refresh()
        return [x for x in self._names.keys()]

    def ids(self):
        if not self.data:
            self.refresh()
        return [x for x in self._jamf_ids.keys()]

    def recordWithId(self, x):
        if not self.data:
            self.refresh()
        if type(x) is str:
            x = int(x)
        return self._jamf_ids.get(x)

    def recordWithName(self, x):
        stderr(
            "WARNING: recordWithName deprecated, use recordsWithName (it turns out names are not unique)."
        )
        names = self.recordsWithName(self, x)[0]
        if len(names) > 1:
            stderr(
                "There is more than one record with the name you are searching for! Only the first one is being used."
            )
        return names[0]

    def recordsWithName(self, x):
        if not self.data:
            self.refresh()
        return [self._names.get(x)]

    def recordsWithRegex(self, x):
        if not self.data:
            self.refresh()
        found = []
        for name in self._names:
            if re.search(x, name):
                found.append(self._names[name])
        return found

    def refresh(self):
        p1, p2, id1, p3 = self.s.swagger(self.cls, "p1, p2, id1, p3")
        lst = self.api.get(p1)  # e.g. categories
        if p2 in lst:
            self.data = lst[p2]
            if not self.data or "size" not in self.data or self.data["size"] == "0":
                self._records = {}
                self._names = {}
                self._jamf_ids = {}
            elif p3 in self.data:  # e.g. category
                records = self.data[p3]
                for d in records:
                    c = self.singular_class(d[id1], d["name"])  # e.g. id1 = "id"
                    c.plural_class = self.cls
                    self._records.setdefault(int(d[id1]), c)  # e.g. id1 = "id"
                    self._names.setdefault(c.name, c)
                    self._jamf_ids.setdefault(c.id, c)
            else:
                pprint(self.data)
                raise JamfError(
                    f"Endpoint {p1} - " f"{p2} has no member named " f"{p3} (p3)."
                )
        else:
            raise JamfError(
                f"Endpoint {p1} has no "
                f"member named {p2}. Check "
                f"the swagger definition file for the name of "
                f"{p1} and set the property "
                f"p2 for class ({p1})."
            )

    def createNewRecord(self, args):
        return self.singular_class(0, args)

    def find(self, x):
        if not self.data:
            self.refresh()
        if isinstance(x, int):
            # check for record id
            result = self._jamf_ids.get(x)
        elif isinstance(x, str):
            try:
                result = self._jamf_ids.get(int(x))
            except ValueError:
                result = self._names.get(x)
        elif isinstance(x, dict):
            keys = ("id", "jamf_id", "name")
            try:
                key = [k for k in x.keys() if k in keys][0]
            except IndexError:
                result = None
            else:
                if key in ("id", "jamf_id"):
                    result = self._jamf_ids.get(int(x[key]))
                elif key == "name":
                    result = self._names.get(key)
        elif isinstance(x, Record):
            result = x
        else:
            raise TypeError(f"can't look for {type(x)}")
        return result


class AdvancedComputerSearch(Record):
    plural_class = "AdvancedComputerSearches"


class AdvancedComputerSearches(Records, metaclass=Singleton):
    # http://localhost/computers.html
    singular_class = AdvancedComputerSearch


class AdvancedMobileDeviceSearch(Record):
    plural_class = "AdvancedMobileDeviceSearches"


class AdvancedMobileDeviceSearches(Records, metaclass=Singleton):
    # http://localhost/mobileDevices.html
    singular_class = AdvancedMobileDeviceSearch


class AdvancedUserSearch(Record):
    plural_class = "AdvancedUserSearches"


class AdvancedUserSearches(Records, metaclass=Singleton):
    # http://localhost/users.html
    singular_class = AdvancedUserSearch


class Building(Record):
    plural_class = "Buildings"


class Buildings(Records, metaclass=Singleton):
    # http://localhost/view/settings/network/buildings
    singular_class = Building


class BYOProfile(Record):
    plural_class = "BYOProfiles"


class BYOProfiles(Records, metaclass=Singleton):
    singular_class = BYOProfile


class Category(Record):
    plural_class = "Categories"


class Categories(Records, metaclass=Singleton):
    # http://localhost/categories.html
    singular_class = Category


class Class(Record):
    plural_class = "Classes"


class Classes(Records, metaclass=Singleton):
    # http://localhost/classes.html
    singular_class = Class


class Computer(Record):
    plural_class = "Computers"

    def apps_print_during(self):
        plural_cls = eval(self.cls.plural_class)
        if not hasattr(plural_cls, "app_list"):
            plural_cls.app_list = {}
        if not hasattr(plural_cls, "computers"):
            plural_cls.computers = {}
        plural_cls.computers[self.name] = True
        try:
            apps = self.get_path("software/applications/application/path")
        except NotFound:
            apps = None
        try:
            versions = self.get_path("software/applications/application/version")
        except NotFound:
            versions = None
        if apps:
            for ii, app in enumerate(apps):
                ver = versions[ii]
                if app not in plural_cls.app_list:
                    plural_cls.app_list[app] = {}
                if ver not in plural_cls.app_list[app]:
                    plural_cls.app_list[app][ver] = {}
                plural_cls.app_list[app][ver][self.name] = True


class Computers(Records, metaclass=Singleton):
    # http://localhost/computers.html?queryType=COMPUTERS&query=
    singular_class = Computer
    sub_commands = {
        "apps": {"required_args": 0, "args_description": ""},
    }

    def apps_print_after(self):
        print("application,version,", end="")
        for computer in self.computers:
            print(computer, ",", end="")
        print("")
        for app, versions in self.app_list.items():
            for version, bla in versions.items():
                text = f'"{app}","{version}",'
                print(text, end="")
                for computer in self.computers:
                    if computer in bla:
                        print("X,", end="")
                    else:
                        print(",", end="")
                print("")


class ComputerConfiguration(Record):
    plural_class = "ComputerConfigurations"


class ComputerConfigurations(Records, metaclass=Singleton):
    singular_class = ComputerConfiguration


class ComputerExtensionAttribute(Record):
    plural_class = "ComputerExtensionAttributes"


class ComputerExtensionAttributes(Records, metaclass=Singleton):
    # http://localhost/computerExtensionAttributes.html
    singular_class = ComputerExtensionAttribute


class ComputerGroup(Record):
    plural_class = "ComputerGroups"


class ComputerGroups(Records, metaclass=Singleton):
    # http://localhost/smartComputerGroups.html
    # http://localhost/staticComputerGroups.html
    singular_class = ComputerGroup


class ComputerReport(Record):
    plural_class = "ComputerReports"


class ComputerReports(Records, metaclass=Singleton):
    singular_class = ComputerReport


class Department(Record):
    plural_class = "Departments"


class Departments(Records, metaclass=Singleton):
    # http://localhost/departments.html
    singular_class = Department


class DirectoryBinding(Record):
    plural_class = "DirectoryBindings"


class DirectoryBindings(Records, metaclass=Singleton):
    singular_class = DirectoryBinding


class DiskEncryptionConfiguration(Record):
    plural_class = "DiskEncryptionConfigurations"


class DiskEncryptionConfigurations(Records, metaclass=Singleton):
    # http://localhost/diskEncryptions.html
    singular_class = DiskEncryptionConfiguration


class DistributionPoint(Record):
    plural_class = "DistributionPoints"


class DistributionPoints(Records, metaclass=Singleton):
    singular_class = DistributionPoint


class DockItem(Record):
    plural_class = "DockItems"


class DockItems(Records, metaclass=Singleton):
    # http://localhost/dockItems.html
    singular_class = DockItem


class Ebook(Record):
    plural_class = "Ebooks"


class Ebooks(Records, metaclass=Singleton):
    singular_class = Ebook


class Ibeacon(Record):
    plural_class = "Ibeacons"


class Ibeacons(Records, metaclass=Singleton):
    singular_class = Ibeacon


class JSONWebTokenConfiguration(Record):
    plural_class = "JSONWebTokenConfigurations"


class JSONWebTokenConfigurations(Records, metaclass=Singleton):
    singular_class = JSONWebTokenConfiguration


class LDAPServer(Record):
    plural_class = "LDAPServers"


class LDAPServers(Records, metaclass=Singleton):
    singular_class = LDAPServer


class MacApplication(Record):
    plural_class = "MacApplications"


class MacApplications(Records, metaclass=Singleton):
    singular_class = MacApplication


class ManagedPreferenceProfile(Record):
    plural_class = "ManagedPreferenceProfiles"


class ManagedPreferenceProfiles(Records, metaclass=Singleton):
    singular_class = ManagedPreferenceProfile


class MobileDevice(Record):
    plural_class = "MobileDevices"


class MobileDevices(Records, metaclass=Singleton):
    # http://localhost/mobileDevices.html?queryType=MOBILE_DEVICES&query=
    singular_class = MobileDevice


class MobileDeviceApplication(Record):
    plural_class = "MobileDeviceApplications"


class MobileDeviceApplications(Records, metaclass=Singleton):
    singular_class = MobileDeviceApplication


class MobileDeviceCommand(Record):
    plural_class = "MobileDeviceCommands"


class MobileDeviceCommands(Records, metaclass=Singleton):
    singular_class = MobileDeviceCommand


class MobileDeviceConfigurationProfile(Record):
    plural_class = "MobileDeviceConfigurationProfiles"


class MobileDeviceConfigurationProfiles(Records, metaclass=Singleton):
    singular_class = MobileDeviceConfigurationProfile


class MobileDeviceEnrollmentProfile(Record):
    plural_class = "MobileDeviceEnrollmentProfiles"


class MobileDeviceEnrollmentProfiles(Records, metaclass=Singleton):
    singular_class = MobileDeviceEnrollmentProfile


class MobileDeviceExtensionAttribute(Record):
    plural_class = "MobileDeviceExtensionAttributes"


class MobileDeviceExtensionAttributes(Records, metaclass=Singleton):
    # http://localhost/mobileDeviceExtensionAttributes.html
    singular_class = MobileDeviceExtensionAttribute


class MobileDeviceInvitation(Record):
    plural_class = "MobileDeviceInvitations"


class MobileDeviceInvitations(Records, metaclass=Singleton):
    singular_class = MobileDeviceInvitation


class MobileDeviceProvisioningProfile(Record):
    plural_class = "MobileDeviceProvisioningProfiles"


class MobileDeviceProvisioningProfiles(Records, metaclass=Singleton):
    singular_class = MobileDeviceProvisioningProfile


class NetbootServer(Record):
    plural_class = "NetbootServers"


class NetbootServers(Records, metaclass=Singleton):
    singular_class = NetbootServer


class NetworkSegment(Record):
    plural_class = "NetworkSegments"


class NetworkSegments(Records, metaclass=Singleton):
    singular_class = NetworkSegment


class OSXConfigurationProfile(Record):
    plural_class = "OSXConfigurationProfiles"


class OSXConfigurationProfiles(Records, metaclass=Singleton):
    singular_class = OSXConfigurationProfile


def parse_package_name(name):
    m = re.match(r"([^-]*)-(.*)\.([^\.]*)$", name)
    if m:
        return m[1], m[2], m[3]
    else:
        m = re.match(r"([^-]*)\.([^\.]*)$", name)
        if m:
            return m[1], "", m[2]
        else:
            return name, "", ""


class Package(Record):
    plural_class = "Packages"

    @property
    def metadata(self):
        if not hasattr(self, "meta"):
            basename, version, filetype = parse_package_name(self.name)
            self._metadata = {
                "basename": basename,
                "version": version,
                "filetype": filetype,
            }
            if not self._metadata["basename"] in self.plural.groups:
                self.plural.groups[self._metadata["basename"]] = []
            if self not in self.plural.groups[self._metadata["basename"]]:
                self.plural.groups[self._metadata["basename"]].append(self)

        return self._metadata

    def refresh_related(self):
        related = {}
        patchsoftwaretitles = jamf_records(PatchSoftwareTitles)
        patchsoftwaretitles_definitions = {}
        for title in patchsoftwaretitles:
            pkgs = title.get_path("versions/version/package/name")
            versions = title.get_path("versions/version/software_version")
            if pkgs:
                for ii, pkg in enumerate(pkgs):
                    if pkg is None:
                        continue
                    if not str(title.id) in patchsoftwaretitles_definitions:
                        patchsoftwaretitles_definitions[str(title.id)] = {}
                    patchsoftwaretitles_definitions[str(title.id)][versions[ii]] = pkg
                    if pkg not in related:
                        related[pkg] = {"PatchSoftwareTitles": []}
                    if "PatchSoftwareTitles" not in related[pkg]:
                        related[pkg]["PatchSoftwareTitles"] = []
                    temp = title.name + " - " + versions[ii]
                    related[pkg]["PatchSoftwareTitles"].append(temp)
        patchpolicies = jamf_records(PatchPolicies)
        for policy in patchpolicies:
            patchsoftwaretitle_id = policy.get_path("software_title_configuration_id")
            parent_pkg_version = policy.get_path("general/target_version")
            if str(patchsoftwaretitle_id) in patchsoftwaretitles_definitions:
                patch_definitions = patchsoftwaretitles_definitions[
                    str(patchsoftwaretitle_id)
                ]
                if parent_pkg_version in patch_definitions:
                    pkg = patch_definitions[parent_pkg_version]
                    if pkg not in related:
                        related[pkg] = {"PatchPolicies": []}
                    if "PatchPolicies" not in related[pkg]:
                        related[pkg]["PatchPolicies"] = []
                    related[pkg]["PatchPolicies"].append(policy.name)
        policies = jamf_records(Policies)
        for policy in policies:
            try:
                pkgs = policy.get_path("package_configuration/packages/package/name")
            except NotFound:
                pkgs = []
            if pkgs:
                for pkg in pkgs:
                    if pkg not in related:
                        related[pkg] = {"Policies": []}
                    if "Policies" not in related[pkg]:
                        related[pkg]["Policies"] = []
                    related[pkg]["Policies"].append(policy.name)
        groups = jamf_records(ComputerGroups)
        for group in groups:
            try:
                criterions = group.get_path("criteria/criterion/value")
            except NotFound:
                criterions = None
            if criterions:
                for pkg in criterions:
                    if pkg and re.search(".pkg|.zip|.dmg", pkg[-4:]):
                        if pkg not in related:
                            related[pkg] = {"ComputerGroups": []}
                        if "ComputerGroups" not in related[pkg]:
                            related[pkg]["ComputerGroups"] = []
                        related[pkg]["ComputerGroups"].append(group.name)
        self.__class__._related = related

    @property
    def related(self):
        if not hasattr(self.__class__, "_related"):
            self.refresh_related()
        if self.name in self.__class__._related:
            return self.__class__._related[self.name]
        else:
            return {}

    def usage_print_during(self):
        related = self.related
        if "PatchSoftwareTitles" in related:
            print(self.name)
        else:
            print(f"{self.name} [no patch defined]")
        if "Policies" in related:
            print("  Policies")
            for x in related["Policies"]:
                print("    " + x)
        if "ComputerGroups" in related:
            print("  ComputerGroups")
            for x in related["ComputerGroups"]:
                print("    " + x)
        if "PatchPolicies" in related:
            print("  PatchPolicies")
            for x in related["PatchPolicies"]:
                print("    " + x)
        print()


class Packages(Records, metaclass=Singleton):
    singular_class = Package
    groups = {}
    sub_commands = {
        "usage": {"required_args": 0, "args_description": ""},
    }


class PatchExternalSource(Record):
    plural_class = "PatchExternalSources"


class PatchExternalSources(Records, metaclass=Singleton):
    singular_class = PatchExternalSource


class PatchInternalSource(Record):
    plural_class = "PatchInternalSources"


class PatchInternalSources(Records, metaclass=Singleton):
    singular_class = PatchInternalSource


class PatchPolicy(Record):
    plural_class = "PatchPolicies"

    def set_version_update_during(self, pkg_version):
        change_made = False
        cur_ver = self.data["general"]["target_version"]
        if cur_ver != pkg_version:
            print(f"Set version to {pkg_version}")
            self.data["general"]["target_version"] = pkg_version
            change_made = True
        else:
            print(f"Version is already {pkg_version}")
        return change_made

    def save(self):
        # If you don't delete the self_service_icon then it will error
        #   <Response [409]>: PUT - https://example.com:8443/JSSResource/patchpolicies/id/18:
        #   Conflict: Error: Problem with icon
        #   Couldn't save changed record: <Response [409]>
        if "self_service_icon" in self.data["user_interaction"]:
            del self.data["user_interaction"]["self_service_icon"]
        return super().save()


class PatchPolicies(Records, metaclass=Singleton):
    singular_class = PatchPolicy
    sub_commands = {
        "set_version": {"required_args": 1, "args_description": ""},
    }


class PatchSoftwareTitle(Record):
    plural_class = "PatchSoftwareTitles"

    def packages_print_during(self):
        print(self.name)
        versions = self.data["versions"]["version"]
        if not type(versions) is list:
            versions = [versions]
        for version in versions:
            if version["package"] is not None:
                print(f" {version['software_version']}: {version['package']['name']}")

    def patchpolicies_print_during(self):
        print(self.name)
        patchpolicies = jamf_records(PatchPolicies)
        for policy in patchpolicies:
            try:
                policy_id = policy.get_path("software_title_configuration_id")
            except NotFound:
                policy_id = None
            if str(policy_id) != str(self.id):
                continue
            print(f" {policy.data['general']['target_version']}: {str(policy)}")

    def set_all_packages_update_during(self):
        policy_regex = {
            "1Password 7": r"^1Password-%VERSION%\.pkg",
            "Apple GarageBand 10": r"^GarageBand-%VERSION%\.pkg",
            "Apple Keynote": r"^Keynote-%VERSION%\.pkg",
            "Apple Numbers": r"^Numbers-%VERSION%\.pkg",
            "Apple Pages": r"^Pages-%VERSION%\.pkg",
            "Apple Xcode": r"^Xcode-%VERSION%\.pkg",
            "Apple iMovie": r"^iMovie-%VERSION%\.pkg",
            "Arduino IDE": r"^Arduino-%VERSION%\.pkg",
            "Bare Bones BBEdit": r"BBEdit-%VERSION%\.pkg",
            "BusyCal 3": r"^BusyCal-%VERSION%\.pkg",
            "Microsoft Remote Desktop 10": r"^Microsoft Remote Desktop-%VERSION%\.pkg",
            "Microsoft Visual Studio Code": r"^Visual Studio Code-%VERSION%\.pkg",
            "Microsoft Teams": r"^Microsoft_Teams_%VERSION%\.pkg",
            "Mozilla Firefox": r"^Firefox-%VERSION%\.pkg",
            "R for Statistical Computing": r"^R-%VERSION%\.pkg",
            "RStudio Desktop": r"RStudio-%VERSION%\.dmg",
            "Sublime Text 3": r"Sublime Text-%VERSION%\.pkg",
            "VLC media player": r"VLC-%VERSION%\.pkg",
            "VMware Fusion 12": r"VMware Fusion-%VERSION%\.pkg",
            "VMware Horizon 8 Client": r"VMwareHorizonClient-%VERSION%.pkg",
            "Zoom Client for Meetings": r"Zoom-%VERSION%.pkg",
        }
        change_made = False
        packages = jamf_records(Packages)
        versions = self.data["versions"]["version"]
        if not type(versions) is list:
            versions = [versions]
        for pkg_version in versions:
            for package in packages:
                if not pkg_version["package"]:
                    if self.name in policy_regex:
                        regex = policy_regex[self.name]
                        regex = regex.replace(
                            "%VERSION%", pkg_version["software_version"]
                        )
                    else:
                        regex = (
                            rf".*{self.name}.*{pkg_version['software_version']}\.pkg"
                        )
                    regex = regex.replace("(", "\\(")
                    regex = regex.replace(")", "\\)")
                    if re.search(regex, package.name):
                        print(f"Matched {package.name}")
                        pkg_version["package"] = {"name": package.name}
                        change_made = True
        return change_made

    def set_package_for_version_update_during(self, package, target_version):
        change_made = False
        versions = self.data["versions"]["version"]
        if not type(versions) is list:
            versions = [versions]
        for pkg_version in versions:
            if pkg_version["software_version"] == target_version:
                print(f"{target_version}: {package}")
                pkg_version["package"] = {"name": package}
                change_made = True
        return change_made

    def versions_print_during(self):
        print(self.name)
        versions = self.data["versions"]["version"]
        if not type(versions) is list:
            versions = [versions]
        for version in versions:
            if version["package"] is not None:
                print(f" {version['software_version']}: {version['package']['name']}")
            else:
                print(f" {version['software_version']}: -")


class PatchSoftwareTitles(Records, metaclass=Singleton):
    singular_class = PatchSoftwareTitle
    sub_commands = {
        "patchpolicies": {"required_args": 0, "args_description": ""},
        "packages": {"required_args": 0, "args_description": ""},
        "set_package_for_version": {"required_args": 2, "args_description": ""},
        "set_all_packages": {"required_args": 0, "args_description": ""},
        "versions": {"required_args": 0, "args_description": ""},
    }


class Peripheral(Record):
    plural_class = "Peripherals"


class Peripherals(Records, metaclass=Singleton):
    singular_class = Peripheral


class PeripheralType(Record):
    plural_class = "PeripheralTypes"


class PeripheralTypes(Records, metaclass=Singleton):
    # I have no idea how to view this data in the web interface
    singular_class = PeripheralType


class Policy(Record):
    plural_class = "Policies"

    def spreadsheet_print_during(self):
        print(self.spreadsheet())

    def spreadsheet(self):

        # Name
        _text = f"{self.name}\t"
        # Category
        _text += f"{self.data['general']['category']['name']}\t"
        # Frequency
        _text += f"{self.data['general']['frequency']}\t"

        # Trigger
        _trigger = []
        if self.data["general"]["trigger_other"]:
            _trigger.append(f"{self.data['general']['trigger_other']}")
        if self.data["general"]["trigger_enrollment_complete"] == "true":
            _trigger.append("Enrollment")
        if self.data["general"]["trigger_startup"] == "true":
            _trigger.append("Startup")
        if self.data["general"]["trigger_login"] == "true":
            _trigger.append("Login")
        if (
            hasattr(self.data["general"], "trigger_logout")
            and self.data["general"]["trigger_logout"] == "true"
        ):
            _trigger.append("Logout")
        if self.data["general"]["trigger_network_state_changed"] == "true":
            _trigger.append("Network")
        if self.data["general"]["trigger_checkin"] == "true":
            _trigger.append("Checkin")
        _text += ", ".join(_trigger)
        _text += "\t"

        # Scope
        _scope = []
        if self.data["scope"]["all_computers"] == "true":
            _scope.append("all_computers")
        if self.data["scope"]["buildings"]:
            for a in self.force_array(self.data["scope"]["buildings"], "building"):
                _scope.append(a["name"])
        if self.data["scope"]["computer_groups"]:
            for a in self.force_array(
                self.data["scope"]["computer_groups"], "computer_group"
            ):
                _scope.append(a["name"])
        if self.data["scope"]["computers"]:
            for a in self.force_array(self.data["scope"]["computers"], "computer"):
                _scope.append(a["name"])
        if self.data["scope"]["departments"]:
            for a in self.force_array(self.data["scope"]["departments"], "department"):
                _scope.append(a["name"])
        if self.data["scope"]["limit_to_users"]["user_groups"]:
            _scope.append("limit_to_users")
        _text += ", ".join(_scope)
        _text += "\t"

        # Packages
        if self.data["package_configuration"]["packages"]["size"] != "0":
            for a in self.force_array(
                self.data["package_configuration"]["packages"], "package"
            ):
                _text += a["name"]
        _text += "\t"

        # Printers
        if self.data["printers"]["size"] != "0":
            for a in self.force_array(self.data["printers"], "printer"):
                _text += a["name"]

            _text += self.data["printers"]["size"]
        _text += "\t"

        # Scripts
        if self.data["scripts"]["size"] != "0":
            for a in self.force_array(self.data["scripts"], "script"):
                _text += a["name"]
        _text += "\t"

        # Self Service
        _self_service = []
        if self.data["self_service"]["use_for_self_service"] != "false":
            _self_service.append("Yes")
        if len(_self_service) > 0:
            _text += ", ".join(_self_service)
        _text += "\t"

        ###############

        # Account Maintenance
        if self.data["account_maintenance"]["accounts"]["size"] != "0":
            for a in self.force_array(
                self.data["account_maintenance"]["accounts"], "account"
            ):
                _text += a["username"]
        _text += "\t"

        # Disk Encryption
        if self.data["disk_encryption"]["action"] != "none":
            _text += self.data["disk_encryption"]["action"]
        _text += "\t"

        # Dock Items
        if self.data["dock_items"]["size"] != "0":
            _text += self.data["dock_items"]["size"]
        _text += "\t"

        return _text

    def promote_update_during(self):
        if "package" not in self.data["package_configuration"]["packages"]:
            print(f"{self.name} has no package to update.")
            return True
        my_packages = self.data["package_configuration"]["packages"]["package"]

        all_packages = jamf_records(Packages)
        print(self.name)
        made_change = False
        for my_package in my_packages:
            similar_packages = []
            search_str = re.sub("-.*", "", my_package["name"])
            for package in all_packages:
                if package.name.find(search_str) == 0:
                    similar_packages.append(package.name)
            if len(similar_packages) > 1:
                index = 1
                for similar_package in reversed(similar_packages):
                    if similar_package == my_package["name"]:
                        print(f"  {index}. {similar_package} [Current]")
                    else:
                        print(f"  {index}. {similar_package}")
                    index += 1
                answer = "0"
                choices = list(map(str, range(1, index)))
                choices.append("")
                while answer not in choices:
                    answer = input("Choose a package [return skips]: ")
                if answer == "":
                    return True
                answer = len(similar_packages) - int(answer)
                my_package["name"] = similar_packages[answer]
                del my_package["id"]
                made_change = True
        if made_change:
            pprint(self.data["package_configuration"]["packages"])
            self.save()


class Policies(Records, metaclass=Singleton):
    singular_class = Policy
    sub_commands = {
        "promote": {"required_args": 0, "args_description": ""},
        "spreadsheet": {"required_args": 0, "args_description": ""},
    }

    def spreadsheet_print_before(self):
        header = [
            "Name",
            "Category",
            "Frequency",
            "Trigger",
            "Scope",
            "Packages",
            "Printers",
            "Scripts",
            "Self Service",
            "Account Maintenance",
            "Disk Encryption",
            "Dock Items",
        ]
        print("\t".join(header))


class Printer(Record):
    plural_class = "Printers"


class Printers(Records, metaclass=Singleton):
    # http://localhost/printers.html
    singular_class = Printer


class RemovableMACAddress(Record):
    plural_class = "RemovableMACAddresses"


class RemovableMACAddresses(Records, metaclass=Singleton):
    # I have no idea how to view this data in the web interface
    singular_class = RemovableMACAddress


class Script(Record):
    plural_class = "Scripts"

    def script_contents_print_during(self):
        try:
            printme = self.get_path("script_contents")
            print(printme[0])
        except NotFound:
            printme = None


class Scripts(Records, metaclass=Singleton):
    # http://localhost/view/settings/computer/scripts
    singular_class = Script
    sub_commands = {
        "script_contents": {"required_args": 0, "args_description": ""},
    }


class Site(Record):
    plural_class = "Sites"


class Sites(Records, metaclass=Singleton):
    # http://localhost/sites.html
    singular_class = Site


class SoftwareUpdateServer(Record):
    plural_class = "SoftwareUpdateServers"


class SoftwareUpdateServers(Records, metaclass=Singleton):
    singular_class = SoftwareUpdateServer


class User(Record):
    plural_class = "Users"


class Users(Records, metaclass=Singleton):
    # http://localhost/users.html?query=
    singular_class = User


class UserExtensionAttribute(Record):
    plural_class = "UserExtensionAttributes"


class UserExtensionAttributes(Records, metaclass=Singleton):
    # http://localhost/userExtensionAttributes.html
    singular_class = UserExtensionAttribute


class UserGroup(Record):
    plural_class = "UserGroups"


class UserGroups(Records, metaclass=Singleton):
    singular_class = UserGroup


class VPPAccount(Record):
    plural_class = "VPPAccounts"


class VPPAccounts(Records, metaclass=Singleton):
    singular_class = VPPAccount


class VPPAssignment(Record):
    plural_class = "VPPAssignments"


class VPPAssignments(Records, metaclass=Singleton):
    singular_class = VPPAssignment


class VPPInvitation(Record):
    plural_class = "VPPInvitations"


class VPPInvitations(Records, metaclass=Singleton):
    singular_class = VPPInvitation


class WebHook(Record):
    plural_class = "WebHooks"


class WebHooks(Records, metaclass=Singleton):
    singular_class = WebHook


def jamf_records(cls, name="", exclude=()):
    """
    Get Jamf Records

    :param cls  <str>:       name of class
    :param name  <str>:      name in record['name']
    :param exclude  <iter>:  record['name'] not in exclude

    :returns:  list of dicts: [{'id': jamf_id, 'name': name}, ...]
    """
    # exclude specified records by full name
    included = [c for c in cls() if c.name not in exclude]
    # NOTE: empty string ('') always in all other strings
    return [c for c in included if name in c.name]


def categories(name="", exclude=()):
    """
    Get Jamf Categories

    :param name  <str>:      name in record['name']
    :param exclude  <iter>:  record['name'] not in exclude

    :returns:  list of dicts: [{'id': jamf_id, 'name': name}, ...]
    """
    return jamf_records(Categories, name, exclude)



# -*- coding: utf-8 -*-

"""
Unittests for jamf.package
"""

__author__ = "Sam Forester"
__email__ = "sam.forester@utah.edu"
__copyright__ = "Copyright (c) 2020 University of Utah, Marriott Library"
__license__ = "MIT"
__version__ = "1.0.1"

import pathlib
import shutil
import unittest

from jamf import package

# location for temporary files created with tests
LOCATION = pathlib.Path(__file__).parent
TMPDIR = LOCATION / "tmp" / "package"
DATA = LOCATION / "data" / "package"


def setUpModule():
    """
    One time setup for entire module.
    If exception is raised, no tests in entire module are run.
    """
    # OPTIONAL
    TMPDIR.mkdir(mode=0o755, parents=True, exist_ok=True)
    package.TMPDIR = TMPDIR / "edu.utah.mlib.package.tests"


def tearDownModule():
    """
    One time cleanup for entire module.
    """
    # OPTIONAL
    shutil.rmtree(TMPDIR, ignore_errors=True)


class BaseTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        self.tmpdir = TMPDIR

    def tearDown(self):
        pass


class TestPackageInit(BaseTestCase):
    # pass
    # def setUp(self):
    #     super().setUp()
    #     self.path = DATA / 'pkgs' / 'edu.utah.mlib.package.test.pkg'

    def test_missing_path_fails(self):
        """
        test FileNotFoundError is raised when given a bad path
        """
        path = pathlib.Path("/does/not/exist")
        if not path.exists():
            with self.assertRaises(FileNotFoundError):
                package.Package(path)

    def test_existing_pkg(self):
        """
        test initialization succeeds
        """
        path = DATA / "pkgs" / "edu.utah.mlib.package.test.pkg"
        if path.exists():
            _ = package.Package(path)


class PackageTests(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.path = DATA / "pkgs" / "edu.utah.mlib.package.test.pkg"
        self.pkg = package.Package(self.path)

    def test_sha512_checksum(self):
        """
        test package sha512 checksum hexdigest
        """
        # `openssl sha512 /path/to/edu.utah.mlib.package.test.pkg`
        expected = (
            "137aa99516b7c9e4d9e8e11e7fffbeb706f380886ff"
            "390da0b786053993903acc1190e3c27c1cde054f351"
            "5369e449f33be15c287aeba116d2e9f1a8dc2f5ff9"
        )
        result = self.pkg.sha512
        self.assertEqual(expected, result)

    def test_md5_checksum(self):
        """
        test package md5 checksum hexdigest
        """
        # `openssl md5 /path/to/edu.utah.mlib.package.test.pkg`
        expected = "a37ae984343b3bbcedb7316e0f3cb349"
        result = self.pkg.md5
        self.assertEqual(expected, result)

    def test_path(self):
        """
        test path is expected
        """
        expected = self.path
        result = self.pkg.path
        self.assertEqual(expected, result)

    def test_name(self):
        """
        test package name is basename of file
        """
        expected = self.path.name
        result = self.pkg.name
        self.assertEqual(expected, result)

    def test_info_keys(self):
        """
        test all package info keys are accounted for
        """
        expected = ["contents", "name", "path", "pkginfo"]
        result = sorted([str(x) for x in self.pkg.info.keys()])
        self.assertEqual(expected, result)

    def test_identifier(self):
        """
        test package identifier
        """
        expected = "test.package"
        result = self.pkg.identifier
        self.assertEqual(expected, result)

    def test_version(self):
        """
        test package version
        """
        expected = "1.0.1"
        result = self.pkg.version
        self.assertEqual(expected, result)

    def test_location(self):
        """
        test package install-location
        """
        expected = "/Applications/Test Package"
        result = self.pkg.location
        self.assertEqual(expected, result)

    def test_apps(self):
        """
        test package app
        """
        expected = [
            {
                "CFBundleIdentifier": "edu.utah.mlib.package.test",
                "CFBundleShortVersionString": "1.0.0",
                "CFBundleVersion": "100",
                "LSMinimumSystemVersion": "10.7.5",
                "name": "package.app",
                "path": "/Applications/Test Package/package.app",
            }
        ]
        result = self.pkg.apps
        self.assertEqual(expected, result)

    def test_pkginfo(self):
        """
        test package information
        """
        expected = {
            "auth": "root",
            "format-version": "2",
            "generator-version": "InstallCmds-681 (18G2022)",
            "identifier": "test.package",
            "install-location": "/Applications/Test Package",
            "overwrite-permissions": "true",
            "postinstall-action": "none",
            "relocatable": "false",
            "version": "1.0.1",
        }
        result = self.pkg.info["pkginfo"]
        self.assertEqual(expected, result)

    def test_size(self):
        """
        test package size
        """
        expected = 2573
        result = self.pkg.size
        self.assertEqual(expected, result)

    def test_expand(self):
        """
        test package expansion
        """
        path = self.pkg.expand()
        self.assertTrue(path.exists())

    def test_cleanup(self):
        """
        test package expansion cleanup
        """
        path = self.pkg.expand()
        del self.pkg
        self.assertFalse(path.exists())

    def test_pkg_as_str(self):
        """
        test package as string
        """
        expected = str(self.pkg.path)
        self.assertTrue(str(self.pkg) == expected)


if __name__ == "__main__":
    unittest.main(verbosity=1)



# -*- coding: utf-8 -*-

"""
Configuration for python-jamf
"""

__author__ = "Sam Forester"
__email__ = "sam.forester@utah.edu"
__copyright__ = "Copyright (c) 2020 University of Utah, Marriott Library"
__license__ = "MIT"
__version__ = "1.3.0"

import getpass
import logging
import plistlib
from datetime import datetime
from os import path, remove
from sys import stderr

import keyring

from .exceptions import JamfConfigError

LINUX_PREFS_TILDA = "~/.edu.utah.mlib.jamfutil.plist"
MACOS_PREFS_TILDA = "~/Library/Preferences/edu.utah.mlib.jamfutil.plist"
AUTOPKG_PREFS_TILDA = "~/Library/Preferences/com.github.autopkg.plist"
JAMF_PREFS = "/Library/Preferences/com.jamfsoftware.jamf.plist"
logging.getLogger(__name__).addHandler(logging.NullHandler())

TOKEN_KEY = "python-jamf-token"
EXPIRE_KEY = "python-jamf-expires"


class Config:
    def __init__(
        self,
        config_path=None,
        hostname=None,
        username=None,
        password=None,
        prompt=False,
    ):
        self.log = logging.getLogger(f"{__name__}.{self.__class__.__name__}")
        self.prompt = prompt
        self.hostname = hostname
        self.username = username
        self.password = password
        self.config_path = resolve_config_path(config_path)
        if not self.hostname and not self.username and not self.password:
            self.load()
        # Prompt for any missing prefs
        if not self.hostname:
            if self.prompt:
                self.hostname = prompt_hostname()
            else:
                raise JamfConfigError(
                    "Config failed to obtain a hostname and prompt is off."
                )
        if not self.username:
            if self.prompt:
                self.username = input("Username: ")
            else:
                raise JamfConfigError(
                    "Config failed to obtain a username and prompt is off."
                )
        if not self.password:
            if self.prompt:
                self.password = getpass.getpass()
            else:
                raise JamfConfigError(
                    "Config failed to obtain a password and prompt is off."
                )
        if not self.hostname.startswith("https://") and not self.hostname.startswith(
            "http://"
        ):
            raise JamfConfigError(
                f"Hostname ({self.hostname}) does not start with 'https://' or 'http://'"
            )

    def load(self):
        if path.exists(self.config_path):
            fptr = open(self.config_path, "rb")
            try:
                prefs = plistlib.load(fptr)
            except plistlib.InvalidFileException:
                fptr.close()
                raise JamfConfigError(
                    f"Could not load {self.config_path}, isit plist formatted?"
                )
            fptr.close()
            if "JSSHostname" in prefs:
                if "Credentials" in prefs:
                    cmessage = f"""
ATTENTION
To improve security with storing credentials used with the jctl tool, we have
deprecated the use of a property list file for storing configuration
information and have migrated to use the Python keyring library provides an
easy way to access the system keyring service from python. It can be used with
the macOS Keychain and Linux KWallet.

Please delete the the configuration at {self.config_path} and recreate it using
the "./jamf/setconfig.py" script.
"""
                    raise JamfConfigError(cmessage)
                self.hostname = prefs["JSSHostname"]
                self.username = prefs["Username"]
                self.password = keyring.get_password(self.hostname, self.username)
            elif "JSS_URL" in prefs:
                self.hostname = prefs["JSS_URL"]
                self.username = prefs["API_USERNAME"]
                self.password = prefs["API_PASSWORD"]
            elif "jss_url" in prefs:
                self.hostname = prefs["jss_url"]
                # No auth in that file
        else:
            raise JamfConfigError(f"Config file does not exist: {self.config_path}")

    def save(self):
        keyring.set_password(self.hostname, self.username, self.password)
        data = {"JSSHostname": self.hostname, "Username": self.username}
        self.log.info(f"saving: {self.config_path}")
        fptr = open(self.config_path, "wb")
        plistlib.dump(data, fptr)
        fptr.close()

    def load_token(self):
        self.token = keyring.get_password(self.hostname, TOKEN_KEY)
        expires = keyring.get_password(self.hostname, EXPIRE_KEY)
        self.expired = False
        if self.token and expires:
            try:
                expires = expires[:-1]  # remove the Z because in case there's no "."
                deadline = datetime.strptime(expires.split(".")[0], "%Y-%m-%dT%H:%M:%S")
                if deadline > datetime.utcnow():
                    self.expired = True
            except ValueError as e:
                stderr.write(
                    f"Error getting saved token: {e}\n"
                    f"expire string 1: {expires}\n"
                    f"expire string 2: {expires.split('.')[0]}\n"
                    f"Will try to continue.\n"
                )

    def save_new_token(self, token, expires):
        self.token = token
        self.expires = expires
        keyring.set_password(self.hostname, TOKEN_KEY, self.token)
        keyring.set_password(self.hostname, EXPIRE_KEY, self.expires)

    def revoke_token(self):
        try:
            keyring.delete_password(self.hostname, TOKEN_KEY)
        except:
            stderr.write("Warning: couldn't delete keyring token\n")
        try:
            keyring.delete_password(self.hostname, EXPIRE_KEY)
        except:
            stderr.write("Warning: couldn't delete keyring token expire date\n")

    def reset(self):
        self.revoke_token()
        try:
            keyring.delete_password(self.hostname, self.username)
        except:
            stderr.write("Warning: couldn't delete keyring password\n")
        remove(self.config_path)


def resolve_config_path(config_path=None):
    if not config_path:
        macos_prefs = path.expanduser(MACOS_PREFS_TILDA)
        linux_prefs = path.expanduser(LINUX_PREFS_TILDA)
        autopkg_prefs = path.expanduser(AUTOPKG_PREFS_TILDA)
        if path.exists(macos_prefs):
            config_path = macos_prefs
        elif path.exists(linux_prefs):
            config_path = linux_prefs
        elif path.exists(autopkg_prefs):
            config_path = autopkg_prefs
        elif path.exists(JAMF_PREFS):
            config_path = JAMF_PREFS
        else:
            config_path = macos_prefs
    if config_path[0] == "~":
        config_path = path.expanduser(config_path)
    return config_path


def prompt_hostname():
    valid = False
    while not valid:
        hostname = input("Hostname (don't forget https:// and :8443): ")
        if hostname.startswith("https://") or hostname.startswith("http://"):
            valid = True
    return hostname



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Python Jamf Exceptions
"""

__author__ = "James Reynolds"
__email__ = "reynolds@biology.utah.edu"
__copyright__ = "Copyright (c) 2022 University of Utah"
__license__ = "MIT"
__version__ = "0.1.1"

import html.parser


# pylint: disable=unnecessary-pass
class Error(Exception):
    """just passing through"""

    pass


class JamfConfigError(Error):
    """Error with the config or keyring"""


class JamfNoConnectionError(Error):
    """Error connecting to the server"""


class JamfRecordNotFound(Error):
    """Record not found"""


class JamfConnectionError(Error):
    """Error connecting to the server"""

    def __init__(self, response, message=None):
        self.response = response
        self.message = message

    def __getattr__(self, attr):
        """
        missing attributes fallback on response
        """
        return getattr(self.response, attr)

    def __str__(self):
        rsp = self.response
        return f"{rsp}: {rsp.request.method} - {rsp.url}: {self.message}"


class JamfAuthenticationError(JamfConnectionError):
    """Error connecting to the server"""


# pylint: disable=super-init-not-called
class APIError(Error):
    """Error in our call"""

    def __init__(self, response):
        self.response = response
        err = parse_html_error(response.text)
        self.message = ": ".join(err) or "failed"
        # if response.status_code == 401:
        #     print(
        #         f'{response.url} returned: "401 Unauthorized". Wrong username/password.'
        #     )
        # elif response.status_code == 404:
        #     print(
        #         f'{response.url} is "404 Not Found". Are you sure this is a Jamf Pro server?'
        #     )
        # elif response.status_code == 503:
        #     print(
        #         f'{response.url} returned: "503 Service Unavailable". Maybe the Jamf server is still starting.'
        #     )
        print(
            f"{response}: {response.request.method} - {response.url}: \n{self.message}"
        )

    def __getattr__(self, attr):
        """
        missing attributes fallback on response
        """
        return getattr(self.response, attr)

    def __str__(self):
        rsp = self.response
        return f"{rsp}: {rsp.request.method} - {rsp.url}: {self.message}"


## The following are used for tests only.


# pylint: disable=too-few-public-methods, abstract-method
class _DummyTag:
    """
    Minimal mock implementation of bs4.element.Tag (only has text attribute)

    >>> eg = _DummyTag('some text')
    >>> eg.text
    'some text'
    """

    def __init__(self, text):
        self.text = text


class JSSErrorParser(html.parser.HTMLParser):
    """
    Minimal mock implementation of bs4.BeautifulSoup()

    >>> [t.text for t in JSSErrorParser(html).find_all('p')]
    ['Unauthorized', 'The request requires user authentication',
     'You can get technical details here. {...}']
    """

    def __init__(self, _html):
        super().__init__()
        self._data = {}
        if _html:
            self.feed(_html)

    def find_all(self, tag):
        """
        Minimal mock implemetation of BeautifulSoup(html).find_all(tag)

        :param tag <str>:   html tag
        :returns <list>:    list of _DummyTags
        """
        return self._data.get(tag, [])

    # pylint: disable=attribute-defined-outside-init
    def handle_data(self, data):
        """
        override HTMLParser().handle_data()
            (automatically called during HTMLParser.feed())
        creates _DummyTag with text attribute from data
        """
        self._dummytag = _DummyTag(data)

    def handle_endtag(self, tag):
        """
        override HTMLParser().handle_endtag()
            (automatically called during HTMLParser.feed())
        add _DummyTag object to dictionary based on tag
        """
        # only create new list if one doesn't already exist
        self._data.setdefault(tag, [])
        self._data[tag].append(self._dummytag)


def parse_html_error(error):
    """
    Get meaningful error information from JSS Error response HTML

    :param html <str>:  JSS HTML error text
    :returns <list>:    list of meaningful error strings
    """
    if not error:
        return []
    soup = JSSErrorParser(error)
    # e.g.: ['Unauthorized', 'The request requires user authentication',
    #        'You can get technical details here. (...)']
    # NOTE: get first two <p> tags from HTML error response
    #       3rd <p> is always 'You can get technical details here...'
    return [t.text for t in soup.find_all("p")][0:2]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
"""

import os
import re

__all__ = ("string", "jamf_version_up_to")


def string():
    try:
        with open(os.path.dirname(__file__) + "/VERSION", "r", encoding="utf-8") as fh:
            version = fh.read().strip()
            if version:
                return version
    except:
        pass
    return "0.0.0"


def jamf_version_up_to(min_version):
    full_version = string()
    try:
        m = re.match(r"^([0-9]+\.[0-9]+\.[0-9]+)", full_version)
        min1 = tuple(map(int, (min_version.split("."))))
        cur = tuple(map(int, (m.group(1).split("."))))
        if min1 <= cur:
            return min_version  # Pass
    except AttributeError:
        return full_version  # Fail
    return full_version  # Fail



# -*- coding: utf-8 -*-

"""
XML and JSON data conversion functions
"""

__author__ = "Sam Forester"
__email__ = "sam.forester@utah.edu"
__copyright__ = "Copyright (c) 2019 University of Utah, Marriott Library"
__license__ = "MIT"
__version__ = "0.2.2"

import xml.sax.saxutils
from collections import defaultdict
from xml.etree import cElementTree as ElementTree


class Error(Exception):
    """just passing through"""

    pass


def etree_to_dict(elem):
    """
    converts xml.cElementTree to python dict
    adapted from: https://stackoverflow.com/a/10077069/12020818
    removed attribute support
    """
    result = {elem.tag: None}
    children = list(elem)
    if children:
        defd = defaultdict(list)
        has_size = False
        for child in children:
            if child.tag == "size":
                has_size = True
            dct = etree_to_dict(child)
            for key, val in dct.items():
                defd[key].append(val)
        result = {}
        for key, val in defd.items():
            #             print(f"{key}, {val[0]}, {type(val[0])}, {has_size}")
            is_array = len(val) > 1 or (has_size and type(val[0]) is dict)
            if elem.tag not in result:
                result[elem.tag] = {}
            if is_array:
                result[elem.tag][key] = val
            else:
                result[elem.tag][key] = val[0]
    elif elem.text:
        result[elem.tag] = elem.text.strip()
    return result


def dict_to_xml(data):
    """
    Convert python dict to xml string
    :returns:  xml string
    """
    if isinstance(data, dict):
        xml_str = ""
        for key, value in data.items():
            if value is None:
                xml_str += f"<{key}/>"
            elif isinstance(value, list):
                # if the value is a list, wrap each entry with the key
                for i in value:
                    result = dict_to_xml(i)
                    xml_str += f"<{key}>{result}</{key}>"
            else:
                # otherwise, wrap the entire result
                result = dict_to_xml(value)
                xml_str += f"<{key}>{result}</{key}>"
    elif isinstance(data, list):
        raise Error("unable to properly tag nested lists")
    else:
        # string, boolean, integers, floats, etc
        xml_str = xml.sax.saxutils.escape(f"{data}")
    return xml_str


def xml_to_dict(xml_string):
    """
    Convert xml string to python dict
    :returns:  dict
    """
    root = ElementTree.XML(xml_string)
    return etree_to_dict(root)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Jamf Config
"""

__author__ = "Sam Forester"
__email__ = "sam.forester@utah.edu"
__copyright__ = "Copyright (c) 2020 University of Utah, Marriott Library"
__license__ = "MIT"
__version__ = "1.0.6"


import argparse
import getpass
import logging
import platform
import sys

import jamf


class Parser:
    def __init__(self):
        myplatform = platform.system()
        if myplatform == "Darwin":
            default_pref = jamf.config.MACOS_PREFS_TILDA
        elif myplatform == "Linux":
            default_pref = jamf.config.LINUX_PREFS_TILDA
        self.parser = argparse.ArgumentParser()
        self.parser.add_argument(
            "-H", "--hostname", help="specify hostname (default: prompt)"
        )
        self.parser.add_argument(
            "-u", "--user", help="specify username (default: prompt)"
        )
        self.parser.add_argument(
            "-p", "--passwd", help="specify password (default: prompt)"
        )
        self.parser.add_argument(
            "-r", "--revoke-token", action="store_true", help="Revoke Bearer token"
        )
        self.parser.add_argument(
            "-C",
            "--config",
            dest="path",
            metavar="PATH",
            default=default_pref,
            help=f"specify config file (default {default_pref})",
        )
        self.parser.add_argument(
            "-P",
            "--print",
            action="store_true",
            help="print existing config profile (except password!)",
        )
        self.parser.add_argument(
            "-t",
            "--test",
            action="store_true",
            help="Connect to the Jamf server using the config file",
        )

    def parse(self, argv):
        """
        :param argv:    list of arguments to parse
        :returns:       argparse.NameSpace object
        """
        return self.parser.parse_args(argv)


def test(config_path):
    api = jamf.API(config_path=config_path)
    try:
        print(api.get("accounts"))
        print("Connection successful")
    except SystemExit as error:
        print(f"Connection failed, check your settings\n{error}")


def print_config(config_path):
    conf = jamf.config.Config(prompt=False, explain=True, config_path=config_path)
    print(conf.hostname)
    print(conf.username)
    if conf.password:
        print("Password is set")
    else:
        print("Password is not set")


def revoke_token(config_path):
    api = jamf.API(config_path=config_path)
    api.revoke_token()


def interactive(args, config_path):
    if args.hostname:
        hostname = args.hostname
    else:
        hostname = jamf.config.prompt_hostname()
    if args.user:
        user = args.user
    else:
        user = input("username: ")
    if args.passwd:
        passwd = args.passwd
    else:
        passwd = getpass.getpass()
    conf = jamf.config.Config(
        hostname=hostname,
        username=user,
        password=passwd,
        prompt=False,
        config_path=config_path,
    )
    conf.save()
    print("Test the config by invoking `conf-python-jamf -t`")


def setconfig(argv):
    logger = logging.getLogger(__name__)
    args = Parser().parse(argv)
    logger.debug(f"args: {args!r}")
    if args.path:
        config_path = args.path
    else:
        myplatform = platform.system()
        if myplatform == "Darwin":
            default_pref = jamf.config.MACOS_PREFS_TILDA
        elif myplatform == "Linux":
            default_pref = jamf.config.LINUX_PREFS_TILDA
        config_path = default_pref
    config_path = jamf.config.resolve_config_path(config_path)
    if args.test:
        test(config_path)
    elif args.print:
        print_config(config_path)
    elif args.revoke_token:
        revoke_token(config_path)
    else:
        interactive(args, config_path)


def main():
    fmt = "%(asctime)s: %(levelname)8s: %(name)s - %(funcName)s(): %(message)s"
    logging.basicConfig(level=logging.INFO, format=fmt)
    setconfig(sys.argv[1:])


if __name__ == "__main__":
    main()



import os
import subprocess

import setuptools

jamf_version = (
    subprocess.run(["git", "describe", "--tags"], stdout=subprocess.PIPE)
    .stdout.decode("utf-8")
    .strip()
)
# assert "." in jamf_version

assert os.path.isfile("jamf/version.py")
if jamf_version != "":
    with open("jamf/VERSION", "w", encoding="utf-8") as fh:
        fh.write(f"{jamf_version}\n")

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="python-jamf",
    version=jamf_version,
    author="The University of Utah",
    author_email="mlib-its-mac@lists.utah.edu",
    description="Python wrapper for Jamf Pro API",
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/univ-of-utah-marriott-library-apple/python-jamf",
    packages=setuptools.find_packages(),
    package_data={"jamf": ["*.json", "VERSION"]},
    include_package_data=True,
    entry_points={"console_scripts": ["conf-python-jamf=jamf.setconfig:main"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        # 4 - Beta
        # 5 - Production/Stable
        "Development Status :: 4 - Beta",
    ],
    python_requires=">=3.6",
    install_requires=["requests>=2.24.0", "keyring>=23.0.0"],
)



# -*- coding: utf-8 -*-
# pylint: disable=relative-beyond-top-level, too-few-public-methods, unused-argument
# pylint: disable=missing-class-docstring, missing-module-docstring, invalid-name

import unittest

from jamf import convert

patchpolicies = """<?xml version="1.0" encoding="UTF-8"?>
<patch_available_titles>
  <size>47</size>
  <available_titles>
    <available_title>
      <name_id>AmaDeusPro</name_id>
      <app_name>Amadeus Pro</app_name>
      <last_modified>2019-08-26T19:50:44.000Z</last_modified>
      <current_version>2.6.1</current_version>
      <publisher>HairerSoft</publisher>
    </available_title>
    <available_title>
      <name_id>Atom</name_id>
      <app_name>Atom</app_name>
      <last_modified>2019-08-21T21:04:28.000Z</last_modified>
      <current_version>1.40.1</current_version>
      <publisher>GitHub</publisher>
    </available_title>
    <available_title>
      <name_id>Audacity</name_id>
      <app_name>Audacity</app_name>
      <last_modified>2019-08-29T17:08:41.000Z</last_modified>
      <current_version>2.3.2.0</current_version>
      <publisher>Audacity</publisher>
    </available_title>
    <available_title>
      <name_id>BBEdit</name_id>
      <app_name>BBEdit</app_name>
      <last_modified>2019-08-06T23:47:01.000Z</last_modified>
      <current_version>12.6.6</current_version>
      <publisher>Barebones</publisher>
    </available_title>
    <available_title>
      <name_id>BoxSync</name_id>
      <app_name>Box Sync</app_name>
      <last_modified>2019-09-09T21:51:18.000Z</last_modified>
      <current_version>4.0.7929</current_version>
      <publisher>Box</publisher>
    </available_title>
    <available_title>
      <name_id>BusyCal</name_id>
      <app_name>BusyCal</app_name>
      <last_modified>2019-09-10T17:42:00.000Z</last_modified>
      <current_version>3.7.0</current_version>
      <publisher>Busy Apps</publisher>
    </available_title>
    <available_title>
      <name_id>CheatSheet</name_id>
      <app_name>CheatSheet</app_name>
      <last_modified>2019-08-22T20:09:41.000Z</last_modified>
      <current_version>1.3.3</current_version>
      <publisher>Media Atelier</publisher>
    </available_title>
    <available_title>
      <name_id>CiscoAnyConnect</name_id>
      <app_name>Cisco AnyConnect</app_name>
      <last_modified>2019-08-22T14:31:31.000Z</last_modified>
      <current_version>4.6.04054</current_version>
      <publisher>Cisco</publisher>
    </available_title>
    <available_title>
      <name_id>CitrixReceiver</name_id>
      <app_name>Citrix Receiver</app_name>
      <last_modified>2019-08-23T15:47:51.000Z</last_modified>
      <current_version>12.9.1</current_version>
      <publisher>Citrix</publisher>
    </available_title>
    <available_title>
      <name_id>Coda2</name_id>
      <app_name>Coda 2</app_name>
      <last_modified>2019-09-09T21:47:28.000Z</last_modified>
      <current_version>2.7.3</current_version>
      <publisher>Panic, Inc.</publisher>
    </available_title>
    <available_title>
      <name_id>Colossus</name_id>
      <app_name>Colossus</app_name>
      <last_modified>2019-08-26T17:40:48.000Z</last_modified>
      <current_version>1.2</current_version>
      <publisher>Sparkfield</publisher>
    </available_title>
    <available_title>
      <name_id>Compressor</name_id>
      <app_name>Compressor</app_name>
      <last_modified>2019-08-27T16:19:02.000Z</last_modified>
      <current_version>4.4.4</current_version>
      <publisher>Apple</publisher>
    </available_title>
    <available_title>
      <name_id>Dropbox</name_id>
      <app_name>Dropbox</app_name>
      <last_modified>2019-09-09T20:41:28.000Z</last_modified>
      <current_version>80.4.127</current_version>
      <publisher>Dropbox, Inc.</publisher>
    </available_title>
    <available_title>
      <name_id>Fetch</name_id>
      <app_name>Fetch</app_name>
      <last_modified>2019-08-26T14:55:55.000Z</last_modified>
      <current_version>5.7.7</current_version>
      <publisher>Fetch Softworks</publisher>
    </available_title>
    <available_title>
      <name_id>FileZilla</name_id>
      <app_name>FileZilla</app_name>
      <last_modified>2019-08-22T15:12:43.000Z</last_modified>
      <current_version>3.44.2</current_version>
      <publisher>Tim Kosse</publisher>
    </available_title>
    <available_title>
      <name_id>GitHubDesktop</name_id>
      <app_name>GitHub Desktop</app_name>
      <last_modified>2019-09-03T16:50:19.000Z</last_modified>
      <current_version>2.1.3</current_version>
      <publisher>GitHub, Inc.</publisher>
    </available_title>
    <available_title>
      <name_id>GitKraken</name_id>
      <app_name>GitKraken</app_name>
      <last_modified>2019-08-27T16:10:50.000Z</last_modified>
      <current_version>6.1.4</current_version>
      <publisher>Axosoft, LLC</publisher>
    </available_title>
    <available_title>
      <name_id>GoogleChrome</name_id>
      <app_name>Google Chrome</app_name>
      <last_modified>2019-08-29T18:40:41.000Z</last_modified>
      <current_version>81</current_version>
      <publisher>Google</publisher>
    </available_title>
    <available_title>
      <name_id>Grammarly</name_id>
      <app_name>Grammarly</app_name>
      <last_modified>2019-09-10T18:34:17.000Z</last_modified>
      <current_version>1.5.52</current_version>
      <publisher>Grammarly, Inc.</publisher>
    </available_title>
    <available_title>
      <name_id>HandBrake</name_id>
      <app_name>HandBrake</app_name>
      <last_modified>2019-09-09T16:01:57.000Z</last_modified>
      <current_version>1.2.2</current_version>
      <publisher>The HandBrake Team</publisher>
    </available_title>
    <available_title>
      <name_id>JamfAdmin</name_id>
      <app_name>Jamf Admin</app_name>
      <last_modified>2017-11-15T18:41:49.000Z</last_modified>
      <current_version>10.0.0</current_version>
      <publisher>Jamf</publisher>
    </available_title>
    <available_title>
      <name_id>JavaSEDevelopmentKit8</name_id>
      <app_name>Java SE Development Kit 8</app_name>
      <last_modified>2017-10-19T12:38:58.000Z</last_modified>
      <current_version>1.8.152</current_version>
      <publisher>Oracle</publisher>
    </available_title>
    <available_title>
      <name_id>KeyboardMaestro</name_id>
      <app_name>Keyboard Maestro</app_name>
      <last_modified>2019-09-16T20:58:49.000Z</last_modified>
      <current_version>9.0.2</current_version>
      <publisher>Stairways Software</publisher>
    </available_title>
    <available_title>
      <name_id>Keynote</name_id>
      <app_name>Keynote</app_name>
      <last_modified>2019-09-06T21:23:23.000Z</last_modified>
      <current_version>9.1</current_version>
      <publisher>Apple, Inc.</publisher>
    </available_title>
    <available_title>
      <name_id>Kindle</name_id>
      <app_name>Kindle</app_name>
      <last_modified>2019-09-10T18:05:41.000Z</last_modified>
      <current_version>1.26.1</current_version>
      <publisher>Amazon.com, Inc.</publisher>
    </available_title>
    <available_title>
      <name_id>LibreOffice</name_id>
      <app_name>LibreOffice</app_name>
      <last_modified>2019-09-06T18:11:02.000Z</last_modified>
      <current_version>6.3.1002</current_version>
      <publisher>The Document Foundation</publisher>
    </available_title>
    <available_title>
      <name_id>MicrosoftRemoteDesktop</name_id>
      <app_name>Microsoft Remote Desktop</app_name>
      <last_modified>2019-09-09T17:30:34.000Z</last_modified>
      <current_version>10.3.2</current_version>
      <publisher>Microsoft</publisher>
    </available_title>
    <available_title>
      <name_id>MicrosoftTeams</name_id>
      <app_name>Microsoft Teams</app_name>
      <last_modified>2019-08-15T19:51:00.000Z</last_modified>
      <current_version>1.00.217856</current_version>
      <publisher>Microsoft</publisher>
    </available_title>
    <available_title>
      <name_id>MozillaFirefox</name_id>
      <app_name>Mozilla Firefox</app_name>
      <last_modified>2019-09-06T16:23:52.000Z</last_modified>
      <current_version>69.0</current_version>
      <publisher>Mozilla</publisher>
    </available_title>
    <available_title>
      <name_id>Numbers</name_id>
      <app_name>Numbers</app_name>
      <last_modified>2019-09-06T21:42:49.000Z</last_modified>
      <current_version>6.1</current_version>
      <publisher>Apple, Inc.</publisher>
    </available_title>
    <available_title>
      <name_id>NVivo</name_id>
      <app_name>NVivo</app_name>
      <last_modified>2019-09-16T20:43:53.000Z</last_modified>
      <current_version>12.2.0</current_version>
      <publisher>QSR International Pty Ltd.</publisher>
    </available_title>
    <available_title>
      <name_id>Opera</name_id>
      <app_name>Opera</app_name>
      <last_modified>2019-08-29T18:35:51.000Z</last_modified>
      <current_version>63.0</current_version>
      <publisher>Opera Software</publisher>
    </available_title>
    <available_title>
      <name_id>Pacifist</name_id>
      <app_name>Pacifist</app_name>
      <last_modified>2019-09-09T16:04:59.000Z</last_modified>
      <current_version>3.6.1</current_version>
      <publisher>CharlesSoft</publisher>
    </available_title>
    <available_title>
      <name_id>Pages</name_id>
      <app_name>Pages</app_name>
      <last_modified>2019-09-06T21:29:21.000Z</last_modified>
      <current_version>8.1</current_version>
      <publisher>Apple, Inc.</publisher>
    </available_title>
    <available_title>
      <name_id>ProfileCreator</name_id>
      <app_name>Profile Creator</app_name>
      <last_modified>2019-09-09T16:07:51.000Z</last_modified>
      <current_version>0.3.2</current_version>
      <publisher>Erik Berglund</publisher>
    </available_title>
    <available_title>
      <name_id>PyCharmCE</name_id>
      <app_name>PyCharm CE</app_name>
      <last_modified>2019-08-30T20:44:35.000Z</last_modified>
      <current_version>2019.2.1</current_version>
      <publisher>JetBrains</publisher>
    </available_title>
    <available_title>
      <name_id>Sketch</name_id>
      <app_name>Sketch</app_name>
      <last_modified>2019-08-22T18:40:46.000Z</last_modified>
      <current_version>57.1</current_version>
      <publisher>Sketch</publisher>
    </available_title>
    <available_title>
      <name_id>Slack</name_id>
      <app_name>Slack</app_name>
      <last_modified>2019-09-06T16:36:46.000Z</last_modified>
      <current_version>4.0.3</current_version>
      <publisher>Slack</publisher>
    </available_title>
    <available_title>
      <name_id>Spark</name_id>
      <app_name>Spark</app_name>
      <last_modified>2019-08-26T21:34:33.000Z</last_modified>
      <current_version>2.3.10</current_version>
      <publisher>Readdle</publisher>
    </available_title>
    <available_title>
      <name_id>Spotify</name_id>
      <app_name>Spotify</app_name>
      <last_modified>2019-09-09T16:18:29.000Z</last_modified>
      <current_version>1.1.10.540</current_version>
      <publisher>Spotify</publisher>
    </available_title>
    <available_title>
      <name_id>SuspiciousPackage</name_id>
      <app_name>Suspicious Package</app_name>
      <last_modified>2019-09-10T18:52:32.000Z</last_modified>
      <current_version>3.5.1</current_version>
      <publisher>Randy Saldinger</publisher>
    </available_title>
    <available_title>
      <name_id>Switch</name_id>
      <app_name>Switch</app_name>
      <last_modified>2019-09-05T21:00:03.000Z</last_modified>
      <current_version>6.50</current_version>
      <publisher>NCH Software</publisher>
    </available_title>
    <available_title>
      <name_id>TableauPublic</name_id>
      <app_name>Tableu Public</app_name>
      <last_modified>2019-09-09T20:37:48.000Z</last_modified>
      <current_version>2019.2.3</current_version>
      <publisher>Tableau Software</publisher>
    </available_title>
    <available_title>
      <name_id>Telegram</name_id>
      <app_name>Telegram</app_name>
      <last_modified>2019-09-09T20:19:19.000Z</last_modified>
      <current_version>5.7</current_version>
      <publisher>Telegram</publisher>
    </available_title>
    <available_title>
      <name_id>Trello</name_id>
      <app_name>Trello</app_name>
      <last_modified>2019-08-26T20:16:21.000Z</last_modified>
      <current_version>2.10.14</current_version>
      <publisher>Atlassian Inc.</publisher>
    </available_title>
    <available_title>
      <name_id>VueScan</name_id>
      <app_name>VueScan</app_name>
      <last_modified>2019-08-15T17:51:50.000Z</last_modified>
      <current_version>9.4.46</current_version>
      <publisher>Hamrick Software</publisher>
    </available_title>
    <available_title>
      <name_id>Zotero</name_id>
      <app_name>Zotero</app_name>
      <last_modified>2019-09-05T23:58:51.000Z</last_modified>
      <current_version>5.0.74</current_version>
      <publisher>Zotero</publisher>
    </available_title>
  </available_titles>
</patch_available_titles>
"""


class ConversionTest(unittest.TestCase):
    def setUp(self):
        self.xml = "<nothing/>"
        self.data = {"nothing": None}

    def test_xml_to_dict(self):
        """
        test conversion of xml string to dict
        """
        expected = self.data
        result = convert.xml_to_dict(self.xml)
        self.assertEqual(expected, result)

    def test_dict_to_xml(self):
        """
        test conversion of dict to xml string
        """
        expected = self.xml
        result = convert.dict_to_xml(self.data)
        self.assertEqual(expected, result)

    def test_dict_reconvert(self):
        """
        test dict -> xml -> dict
        """
        _xml = convert.dict_to_xml(self.data)
        result = convert.xml_to_dict(_xml)
        expected = self.data
        self.assertEqual(expected, result)

    def test_xml_reconvert(self):
        """
        test xml -> dict -> xml
        """
        _dict = convert.xml_to_dict(self.xml)
        result = convert.dict_to_xml(_dict)
        expected = self.xml
        self.assertEqual(expected, result)


class TestSimpleDict(ConversionTest):
    def setUp(self):
        self.xml = "<test><key>value</key></test>"
        self.data = {"test": {"key": "value"}}


class TestSimpleList(ConversionTest):
    def setUp(self):
        self.xml = (
            "<list>"
            "<item>one</item>"
            "<item>two</item>"
            "<item>three</item>"
            "</list>"
        )
        self.data = {"list": {"item": ["one", "two", "three"]}}


class TestListOfDicts(ConversionTest):
    def setUp(self):
        self.xml = (
            "<list>"
            "<item>"
            "<id>1</id>"
            "<name>one</name>"
            "</item>"
            "<item>"
            "<id>2</id>"
            "<name>two</name>"
            "</item>"
            "<item>"
            "<id>3</id>"
            "<name>three</name>"
            "</item>"
            "</list>"
        )
        self.data = {
            "list": {
                "item": [
                    {"id": "1", "name": "one"},
                    {"id": "2", "name": "two"},
                    {"id": "3", "name": "three"},
                ]
            }
        }


@unittest.skip
class TestPatchSoftwareTitle(ConversionTest):
    """
    Too difficult to test

    We can't guarantee the ordering of strings when converting
    from a dictionary to xml string, so even though the values are identical,
    string comparison fails
    """

    def setUp(self):
        self.maxDiff = None
        self.xml = (
            "<patch_software_title>"
            "<id>31</id>"
            "<name>Mozilla Firefox</name>"
            "<name_id>MozillaFirefox</name_id>"
            "<source_id>2</source_id>"
            "<notifications>"
            "<email_notification>true</email_notification>"
            "<web_notification>true</web_notification>"
            "</notifications>"
            "<category>"
            "<id>1</id>"
            "<name>Apps - Web Browsers</name>"
            "</category>"
            "<site>"
            "<id>-1</id>"
            "<name>None</name>"
            "</site>"
            "<versions>"
            "<version>"
            "<software_version>69.0.1</software_version>"
            "<package>"
            "<id>284</id>"
            "<name>firefox_69.0.1_2019.09.18_rcg.pkg</name>"
            "</package>"
            "</version>"
            "<version>"
            "<software_version>69.0</software_version>"
            "<package>"
            "<id>253</id>"
            "<name>firefox_69.0_2019.09.04_rcg.pkg</name>"
            "</package>"
            "</version>"
            "<version>"
            "<software_version>68.0.2</software_version>"
            "<package>"
            "<id>182</id>"
            "<name>firefox_68.0.2_2019.08.20_rcg.pkg</name>"
            "</package>"
            "</version>"
            "<version>"
            "<software_version>68.0.1</software_version>"
            "<package>"
            "<id>121</id>"
            "<name>firefox_68.0.1_2019.07.22_rcg.pkg</name>"
            "</package>"
            "</version>"
            "</versions>"
            "</patch_software_title>"
        )
        self.data = {
            "patch_software_title": {
                "category": {"id": "1", "name": "Apps - Web Browsers"},
                "id": "31",
                "name": "Mozilla Firefox",
                "name_id": "MozillaFirefox",
                "notifications": {
                    "email_notification": "true",
                    "web_notification": "true",
                },
                "site": {"id": "-1", "name": "None"},
                "source_id": "2",
                "versions": {
                    "version": [
                        {
                            "package": {
                                "id": "284",
                                "name": "firefox_69.0.1_2019.09.18_rcg.pkg",
                            },
                            "software_version": "69.0.1",
                        },
                        {
                            "package": {
                                "id": "253",
                                "name": "firefox_69.0_2019.09.04_rcg.pkg",
                            },
                            "software_version": "69.0",
                        },
                        {
                            "package": {
                                "id": "182",
                                "name": "firefox_68.0.2_2019.08.20_rcg.pkg",
                            },
                            "software_version": "68.0.2",
                        },
                        {
                            "package": {
                                "id": "121",
                                "name": "firefox_68.0.1_2019.07.22_rcg.pkg",
                            },
                            "software_version": "68.0.1",
                        },
                    ]
                },
            }
        }

    @unittest.skip("key ordering causes incorrect failure")
    def test_dict_to_xml(self):
        pass


if __name__ == "__main__":
    unittest.main(verbosity=1)



# -*- coding: utf-8 -*-

"""
Tools for interacting with Jamf Admin (the undocumented Jamf API)

See https://apple.lib.utah.edu/reverse-engineering-jamf-admin/
"""

__author__ = "Sam Forester"
__email__ = "sam.forester@utah.edu"
__copyright__ = "Copyright (c) 2020 University of Utah, Marriott Library"
__license__ = "MIT"
__version__ = "1.3.1"

import glob
import hashlib
import json
import logging
import pathlib
import re
import shutil
import subprocess
import sys
import time
import urllib

import requests

from . import config, convert, package, records

# GLOBALS
logger = logging.getLogger(__name__)


class Error(Exception):
    pass


class UploadError(Error):
    pass


class DuplicatePackageError(Error):
    pass


class MissingPackageError(Error):
    pass


class TimeoutError(Error):
    pass


class Singleton(type):
    """
    Use single instance of class
    """

    _instances = {}

    def __call__(cls, *a, **kw):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*a, **kw)
        return cls._instances[cls]


class JamfAdmin(metaclass=Singleton):
    """
    Class for uploading/updating packages in Jamf Admin.app
    """

    def __init__(
        self, config_path=None, hostname=None, username=None, password=None, prompt=True
    ):
        """
        :param hostname <str>:  JSS hostname (e.g. 'your.jss.domain')
        :param auth <tuple>:    JSS authentication credentials (user, passwd)
        :param path <str>:      path to jamf.config.PREFERENCES file
        :param port <int>:      JSS port (default: 8443)
        """
        self.log = logging.getLogger(f"{__name__}.JamfAdmin")

        conf = config.Config(
            config_path=config_path,
            hostname=hostname,
            username=username,
            password=password,
            prompt=prompt,
        )
        hostname = hostname or conf.hostname
        username = username or conf.username
        password = password or conf.password

        if not hostname and not username and not password:
            print("No jamf hostname or credentials could be found.")
            exit(1)

        self.hostname = hostname

        self.session = requests.Session()
        self.session.auth = (username, password)
        self.data = self.authenticate()
        self.categories = records.Categories()
        for share in self.fileservers:
            if share["master"] == "true":
                self.fileshare = FileShare.fromJXML(share)
                break
        self._packages = None
        self._refreshed = {"packages": False}

    def __del__(self):
        for pkg in self.packages:
            if not pkg.path.exists():
                self.log.info(f"removing deleted package record: {pkg.name}")
                self.remove(pkg)

    @property
    def packages(self):
        """
        :returns: list of packages from Jamf Admin
        """
        _trigger = self._refreshed["packages"]
        if not self._packages or _trigger:
            if not self.fileshare.mounted:
                self.fileshare.mount()
            self._refreshed["packages"] = False
            _data = self.data["packages"]["package"]
            # fix xml_to_dict issues by ensuring sure it's a list
            if isinstance(_data, dict):
                _data = [_data]
            self._packages = PackageList()
            for d in _data:
                try:
                    pkg = Package(d, self.fileshare)
                except FileNotFoundError:
                    self.remove(d["id"])
                else:
                    if pkg in self._packages:
                        err = f"duplicate package: {pkg.jssid}: {pkg.name}"
                        self.log.error(err)
                        self.remove(pkg.jssid)
                    else:
                        self._packages.append(pkg)
        return self._packages

    @property
    def fileservers(self):
        """
        :returns: list of fileservers from Jamf Admin
        """
        _data = self.data["fileservers"]["fileserver"]
        # fix xml_to_dict issues by making sure we return a list
        return _data if isinstance(_data, list) else [_data]

    @property
    def groups(self):
        """
        :returns: list of groups from Jamf Admin
        """
        _data = self.data["groups"]["group"]
        # fix xml_to_dict issues by making sure we return a list
        return _data if isinstance(_data, list) else [_data]

    def authenticate(self):
        """
        Authenticate Jamf Admin session
        """
        self.log.debug("authenticating session: %r", self.hostname)
        username, passwd = self.session.auth
        form = {
            "username": username,
            "password": passwd,
            "casperAdminVersion": "",
            "skipComputers": "true",
        }
        # authenticate session (uses 'JSESSIONID' cookie for future requests)
        response = self.session.post(f"{self.hostname}//casper.jxml", data=form)
        # NOTE: failed requests will still return <Response [200]>
        #       failed requests will only have
        if not response.ok:
            # NOTE: response.ok requires requests.__version__ >= '2.18.4'
            # failed requests will still return <Response [200]>
            raise Error(f"{self.hostname}: authentication failed: {response}")
        # convert xml response to dict
        info = convert.xml_to_dict(response.text)
        # check authentication
        # TO-DO: figure out best way to determine bad authentication
        # NOTE: failed requests will only have 'epoch' and 'response' tags
        #       successful requests will not have 'response' tag
        failure = info["jamfServlet"].get("response")
        if failure:
            raise Error(f"{self.hostname}: authentication failed: {failure}")
        return info["jamfServlet"]

    def refresh(self):
        """
        reauthenticate and refresh data
        """
        self.data = self.authenticate()
        self._refreshed = {k: True for k in self._refreshed.keys()}
        return self.data

    def index(self, pkg):
        """
        Not Implemented
        """
        # if not isinstance(pkg, ServerPackage):
        #     raise Error("can't index local package")
        # url = f"{self.url}//packageIndex.jxml"
        raise NotImplementedError()

    def uploaded(self, pkg):
        return self.fileshare.packages in pkg.path.parents

    def upload(self, pkg, verify=True, force=False):
        """
        Upload local package
        """
        if not isinstance(pkg, package.Package):
            raise TypeError(f"invalid package: {pkg!r}")
        # check if package is already uploaded
        if self.uploaded(pkg) and not force:
            raise ValueError(f"package already uploaded: {pkg.name}")
        self.log.info(f"uploading package: {pkg.name}")
        # mount fileshare
        self.fileshare.mount()
        # upload package file
        dest = self.fileshare.packages / pkg.name
        self.log.debug(f"> cp {pkg.path} {dest}")
        subprocess.check_call(["/bin/cp", "-p", pkg.path, dest])
        self.log.debug(f"successfully copied package: {pkg.path}")
        uploaded = package.Package(dest)
        try:
            existing = self.find(uploaded.name)
        except MissingPackageError:
            pass
        else:
            uploaded = existing
        if verify:
            if not self.verify_upload(pkg, uploaded):
                raise UploadError(f"failed to upload package: {pkg.path}")
        return uploaded

    def verify_upload(self, orig, new):
        # verify file copied
        # TO-DO: get size of server file to compare to original (fast)
        # TO-DO: calculate checksums of each (slow)
        self.log.info("verifying package upload")
        self.log.warning("upload verification minimally implemented")
        verified = True
        self.log.debug("checking sizes")
        if orig.size != new.size:
            self.log.error("mismatching package sizes")
            verified = False
        return verified

    def add(self, pkg, force=False):
        # create new package entry in the database
        if not isinstance(pkg, package.Package):
            raise ValueError(f"invalid package: {pkg!r}")
        # check if package was already uploaded
        try:
            uploaded = self.find(pkg.name)
        except MissingPackageError:
            # this is what we want to happen
            uploaded = self.upload(pkg)
        else:
            if not force:
                raise DuplicatePackageError(f"package already added: {pkg}")
            self.log.debug("forcing upload")
            self.upload(pkg)
            return uploaded

        # get package data form
        form = package_upload_form(pkg)
        form.update(
            {"username": self.session.auth[0], "password": self.session.auth[1]}
        )
        # submit package upload form
        self.log.debug("submitting package form")
        url = f"{self.hostname}//casperAdminAddObject.jxml"
        r = self.session.post(url, data=form)
        # get the ID of the newly created package
        jssid = convert.xml_to_dict(r.text)["jamfServlet"]["new_id"]
        logger.debug(f"new package ID: {jssid}")
        added = Package.fromPackage(jssid, pkg, self.fileshare)
        self.packages.append(added)
        self.log.info(f"successfully added: {added.name}")
        return added

    def delete(self, pkg):
        if not isinstance(pkg, Package):
            pkg = self.find(pkg.name)
        self.remove(pkg)
        self.packages.remove(pkg)
        if pkg.path.exists():
            pkg.path.unlink()

    def remove(self, x):
        self.log.info(f"removing package record: {x}")
        if isinstance(x, package.Package):
            jssid = self.find(x.name).jssid
        elif isinstance(x, (str, int)):
            jssid = x
        form = {
            "username": self.session.auth[0],
            "allScriptsMigratedToJSS": "true",
            "deletedPackageID": jssid,
        }
        self.session.post(f"{self.hostname}//casperAdminSave.jxml", data=form)

    def update(self, pkg, notes=""):
        if not hasattr(pkg, "jssid"):
            try:
                pkg = self.find(pkg.name)
            except AttributeError:
                raise TypeError(f"invalid package: {pkg!r}")
            except FileNotFoundError:
                raise TypeError(f"package never added: {pkg!r}")

        self.log.info(f"updating package: id: {pkg.jssid}: {pkg.name}")
        form = package_update_form(pkg)
        form.update(
            {
                "username": self.session.auth[0],
                "password": self.session.auth[1],
                "allScriptsMigratedToJSS": "true",
            }
        )
        if notes:
            form["packageNotes"] = notes
        self.session.post(f"{self.hostname}//casperAdminSave.jxml", data=form)

    def find(self, name):
        """
        :returns: appropriate package for name
        """
        if isinstance(name, package.Package):
            name = name.name
        for pkg in self.packages:
            if pkg.name == name:
                return pkg
        raise MissingPackageError(f"missing package: {name}")


class PackageList(list):
    @property
    def names(self):
        return [p.name for p in self]

    def __contains__(self, x):
        return x.name in self.names


class Package(package.Package):

    _instances = {}

    @classmethod
    def fromJXML(cls, conf, fileshare):
        """
        Create Package object from JamfAdmin info
        """
        try:
            volume = pathlib.Path(fileshare.path)
        except TypeError:
            fileshare.mount()
            volume = pathlib.Path(fileshare.path)
        _ = volume / "Packages" / conf["filename"]
        keys = ("checksum", "hashValue", "info", "notes", "groupid")
        return cls({k: conf[k] for k in keys if k in conf}, fileshare)

    @classmethod
    def fromPackage(cls, jssid, pkg, fileshare):
        """
        Create ServerPackage object from existing Package object
        """
        data = {
            "id": jssid,
            "filename": pkg.name,
            "checksum": pkg.md5,
            "hashValue": pkg.sha512,
        }
        return cls(data, fileshare)

    def __new__(cls, data, fileshare):
        """
        returns existing package if one has been instantiated
        """
        _ = logging.getLogger(__name__)
        jssid = int(data["id"])
        if jssid not in cls._instances:
            cls._instances[jssid] = super(Package, cls).__new__(cls)
        return cls._instances[jssid]

    def __init__(self, data, fileshare):
        path = fileshare.path / "Packages" / data["filename"]
        # if not path.exists():
        #     raise ValueError(f"unable to find package on server: {path}")
        super().__init__(path)
        self.log = logging.getLogger(f"{__name__}.Package")
        self._md5 = data.get("checksum")
        self._sha512 = data.get("hashValue")
        self.jssid = int(data["id"])
        self.category = records.Categories().find(data.get("groupid", "-1"))
        self.notes = data.get("notes", "")
        try:
            self._info = json.loads(data.get("info", "{}"))
        except (json.decoder.JSONDecodeError, TypeError):
            self._info = {}
        try:
            self._min_os_ver = self._info.get("minimum_os")
        except AttributeError:
            pass

    def __eq__(self, x):
        # this significantly speeds up
        return self.name == x.name

    @property
    def info(self):
        _info = super().info
        _info.update({"minimum_os": self.minimum_os})
        return _info


class FileShare:
    @classmethod
    def fromJXML(cls, jxml):
        """
        Use Jamf Admin fileserver info to create new FileShare
        :returns: FileShare object
        """
        protocol = jxml["type"]
        # default domain (fallback to ip), raises KeyError if no fallback
        # NOTE: Jamf Admin incorrectly uses domain in 'ip'
        host = jxml.get("domain") or jxml["ip"]
        auth = (jxml["adminUsername"], jxml["adminPassword"])
        kwargs = {"share": jxml["share"], "name": jxml["displayname"]}
        return cls(protocol, host, auth, **kwargs)

    def __init__(self, protocol, hostname, auth, name=None, share=""):
        self.log = logging.getLogger(f"{__name__}.FileShare")
        self.log.debug(f"protocol: {protocol!r}")
        self.log.debug(f"hostname: {hostname!r}")
        self.log.debug(f"name: {name!r}")
        self.log.debug(f"share: {share!r}")
        self.protocol = protocol
        self.host = hostname
        self.name = name or hostname
        self.auth = auth
        self.share = share
        self.path = None
        # attempt to find previously mounted volume
        if self.mounted:
            # has side-effect of setting self.path (should be changed)
            self.log.info(f"found previously mounted volume: {self.path}")

    # @property
    # def path(self):
    #     if self._path and not self._path.exists():
    #         self._path = None
    #     if not self._path:

    @property
    def mounted(self):
        """
        Check if fileshare is mounted, and if so, update the path
        :returns: True if fileshare is currently mounted
        """
        # look for
        # e.g. 'remote.file.share.domain/ShareVolume'
        # identifier = os.path.join(self.host, self.share)
        identifier = f"{self.host}/{self.share}"
        for device, path, _ in mounted_volumes():
            if identifier in device:
                # update the path of the file share
                self.path = pathlib.Path(path)
                break
        return True if self.path and self.path.exists() else False

    @property
    def packages(self, automount=True):
        """
        :returns <Path>: absolute path to Packages directory on fileshare
        """
        if automount and not self.mounted:
            self.mount()
        # verify there is actually a path to return
        if not self.path:
            raise Error(f"{self.name}: not mounted")
        return self.path / "Packages"

    def _osx_mount(self, user, passwd):
        url = f"{self.protocol}://{user}:{passwd}@{self.host}/{self.share}"
        self.log.debug(f"> open {url.replace(passwd, '******')}")
        subprocess.check_call(["/usr/bin/open", url])

    def _linux_mount(self, user, passwd):
        if self.protocol != "smb":
            raise NotImplementedError(
                f"File upload protocol {self.protocol} not supported on linux machines"
            )
        args = f"-t cifs -o username={user},password={passwd} //{self.host}/{self.share} /mnt/"
        self.log.debug(f"> mount {args.replace(passwd, '******')}")
        subprocess.check_call(["/bin/mount", args])

    def mount(self, timeout=5):
        """
        Mount the fileshare
        :returns <Path>: mounted volume path
        """
        if self.mounted:
            self.log.warning(f"{self.name}: already mounted")
            return self.path

        self.log.debug(f"mounting: {self.name}", self.name)
        user, passwd = [urllib.parse.quote(s) for s in self.auth]
        if sys.platform == "linux" or sys.platform == "linux2":
            self._linux_mount(user, passwd)
        elif sys.platform == "darwin":
            self._osx_mount(user, passwd)
        else:
            raise NotImplementedError(
                f"FileShare.mount not implemented for {sys.platform}"
            )
        self._wait_for_mount(timeout=timeout)
        self.log.info(f"successfully mounted: {self.name}")
        self.log.debug(f"path: {str(self.path)!r}")
        return self.path

    def unmount(self):
        """
        Unmount the fileshare
        """
        if not self.mounted:
            self.log.warning(f"{self.name}: already unmounted")
            return
        self.log.debug(f"unmounting: {self.name}")
        subprocess.check_call(["/sbin/umount", self.path])
        self.path = None
        self.log.info(f"{self.name}: succesfully unmounted")

    def _wait_for_mount(self, timeout=5, poll=1):
        """
        Wait for fileshare to mount
        """
        self.log.debug(f"{self.name}: waiting for mount")
        time.sleep(poll)
        w = poll
        while not self.mounted:
            if w >= timeout:
                self.log.error(f"{self.name}: failed to mount")
                raise TimeoutError(f"timed out after {w} second(s)")
            time.sleep(poll)
            w += poll


def package_upload_form(pkg):
    """
    :returns: form to post for package uploads
    """
    # TO-DO: this will require .mpkg support
    if pkg.path.suffix == ".pkg":
        type_ = "package"
    else:
        raise ValueError(f"unsupported package type: {pkg.path.suffix!r}")
    # return populated form for package
    return {
        "type": type_,
        "packageName": pkg.name,
        "packageFileName": pkg.name,
        "adobeInstall": "false",
        "osInstall": "false",
        "osInstallerVersion": "",
        "parentPackageID": "-1",
        "checksum": str(pkg.md5),
        "hashType": "1",
        "hashValue": str(pkg.sha512),
        "reboot": "false",  # str(pkg.reboot).lower(),
        "packagePriority": "10",
    }


def package_update_form(pkg):
    """
    :returns: form to post for package updates
    """
    # TO-DO: this will require .mpkg support
    if pkg.path.suffix == ".pkg":
        format = "Apple Package"
    else:
        raise ValueError(f"unsupported package type: {pkg.path.suffix!r}")
    try:
        _info = pkg.info
        _info["minimum_os"] = pkg.minimum_os
    except Exception as e:
        _info = {"ERROR": f"failed to dump package info: {e}"}

    try:
        pkginfo = json.dumps(_info, indent=2)
    except Exception:
        pkginfo = ""
    try:
        notes = str(pkg.notes)
    except AttributeError:
        notes = ""
    # return populated form template
    return {
        "packageID": pkg.jssid,
        "packageName": pkg.name,
        "packageFileName": pkg.name,
        "checksum": str(pkg.md5),
        "hashType": "1",
        "hashValue": str(pkg.sha512),
        "packageGroupID": pkg.category.id,
        "packagePriority": "10",
        "packageInfo": pkginfo,
        "packageNotes": notes,
        "packageFormat": format,
        "packageSize": "n/a",
        "packageRequirements": "",
        "bootVolumeRequired": "false",
        "fut": "false",
        "feu": "false",
        "ifswu": "false",
        "reboot": "false",
        "uninstall": "false",
        "packageRequiredProcessor": "None",
        "packageSwitchWithPackageID": "-1",
        "selfHealingAction": "nothing",
        "selfHealingNotify": "false",
        "adobeInstall": "false",
        "adobeInstallerImage": "false",
        "adobeUpdater": "false",
        "osInstall": "false",
        "osInstallerVersion": "",
        "packageSerialNumber": "",
        "parentPackageID": "-1",
        "basePath": "",
        "ignoreConflictingProcesses": "false",
        "suppressFromDock": "false",
        "suppressEULA": "false",
        "suppressRegistration": "false",
        "suppressUpdates": "false",
        "installLanguage": "en_US",
    }


# def package_info(pkg):
#     """
#     :returns <str>:  information about the package
#     """
#     info = (f"identifier: {pkg.identifier}\n"
#             f"version: {pkg.version}\n"
#             f"install-location: {pkg.location}\n"
#             "Contents:")
#     keys = ('path', 'CFBundleIdentifier', 'CFBundleShortVersionString',
#             'CFBundleVersion')
#     for app in pkg.apps:
#         i = "\n".join([f"    {k}: {app[k]}" for k in keys])
#         info += f"\n  {app['name']}:\n{i}"
#     return info


# def package_notes(path):
#     path = pathlib.Path(path)
#     *name, ver, date, author = path.stem.split('_')
#     return f"{date}, {author.upper()}"


def mounted_volumes():
    """
    Uses `/sbin/mount` to get information about mounted file systems
    :returns: list of tuples for each mounted file system
                e.g. [(device, path, types), ...]
    """
    logger = logging.getLogger(__name__)
    # e.g. '/dev/disk1s1 on / (apfs, local, journaled)'
    #  -> ('/dev/disk1s1', '/', 'apfs, local, journaled')
    r = re.compile(r"^(.+) on (.+) \((.+)\)$")
    logger.debug("checking mounted volumes")
    out = subprocess.check_output(["/sbin/mount"]).decode()
    return [re.match(r, x).groups() for x in out.splitlines()]


def package_checksums(path, bufsize=8192):
    """
    calculate md5 and sha512 digests of path
    :param path:        path to file
    :param bufsize:     block size buffer to read (default: 4096)
    :returns:           (md5.hexdigest(), sha512.hexdigest())
    """
    raise NotImplementedError("use jamf.package.calculate_checksums instead")
    logger = logging.getLogger(__name__)
    logger.debug("calculating checksums: %r", path)
    md5 = hashlib.md5()
    sha512 = hashlib.sha512()
    with open(path, "rb") as f:
        while True:
            data = f.read(bufsize)
            if not data:
                break
            md5.update(data)
            sha512.update(data)
    return (md5.hexdigest(), sha512.hexdigest())


def package_index(path, cleanup=True):
    """
    package indexing payload form data for requests.post
    requests.post(url, data=package_index('/path/to/example.pkg'))

    :param  path  <str>:    path of .pkg file
    :returns  <list>:       key/value tuple pairs
    """
    logger = logging.getLogger(__name__)
    # expand package boms (see `man pkgutil` for more info)
    logger.debug(f"> pkgutil --bom {path}")
    boms = subprocess.check_output(["/usr/sbin/pkgutil", "--bom", path])
    # process all boms included in package
    boms = [x for x in boms.splitlines() if x]
    if not boms:
        raise RuntimeError(f"no bill of materials found: {path}")
    # see `man lsbom` for formatting
    logger.debug("> lsbom -p fMguTsc '%s'", "', '".join(boms))
    cmd = ["/usr/bin/lsbom", "-p", "fMguTsc"] + boms
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = p.communicate()

    if p.returncode != 0 or not out:
        raise RuntimeError(err)

    if cleanup:
        # cleanup all bom files created in /tmp by `pkgutil`
        for path in glob.glob(f"/tmp/{path}.boms*"):
            shutil.rmtree(path)

    form = []
    keys = ("path", "mode", "owner", "group", "date", "size", "checksum")
    for line in out.splitlines():
        # list of values per line (converted from byte string)
        values = [x.decode().strip() for x in line.split(b"\t")]

        # NOTE: Jamf Admin removes '.' from beginning of path
        # modify path to exclude leading '.'
        # values[0] = values[0][1:]

        # NOTE: Jamf Admin does not submit package root in its index
        # create entry for every non-root entry in bom (i.e. '.' > 1 == False)
        if len(values[0]) > 1:
            # remove double space between month and single digit dates
            # (e.g. 'Sun Jun  9 01:00:00 2019' -> 'Sun Jun 9 01:00:00 2019')
            values[4] = values[4].replace("  ", " ")
            form.extend([z for z in zip(keys, values)])
    return form
